#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''programs.py: Contains methods for running GSLIB and CCG programs'''
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
import subprocess
import os
import time
import warnings
from ..utility.logging import printerr
from .program_utils import dedent_parstr
from .. pygeostat_parameters import Parameters

# Seed the random number generator for spacing parallel calls
#   and seeding ACORNI random number generators
import random
random.seed()


class Program(object):
    '''Base class containing routines for running GSLIB programs'''

    def __init__(self, program=None, parstr=None, parfile='temp', getpar=None,
                 nogetarg=False, defaultdict={}):
        self.program = self.__check_program(program)
        self.parstr = parstr
        # nogetarg is a flag for FORTRAN programs that call READ
        # instead of using GETARG - this requires that a PIPE be used
        # rather than traditional command line arguments
        self.nogetarg = nogetarg
        self.parfile = parfile
        self.defaultdict = defaultdict
        if getpar is None:
            getpar = Parameters['config.getpar']
        if getpar:
            self.getparfile()

    def __check_program(self, program):

        try:
            popenobject_ = subprocess.Popen(program)
            popenobject_.terminate()
            return program
        except FileNotFoundError:
            if Parameters['config.use_shipped_executable']:
                print('Unable to find the provided program! Trying to use the executable pool shipped with pygeostat!')
                exec_dir = os.path.abspath(os.path.join(os.path.dirname( __file__ ), r'..', r'executable'))
                _, program = os.path.split(program)
                return os.path.join(exec_dir, program)
            else:
                raise FileNotFoundError('The executable file does not exist. \nConsider setting the parameter `config.use_shipped_executable` to True to use the shipped executable files.')


    def run(self, parstr=None, parfile=None, program=None, nogetarg=None,
            filehandle=None, logfile=None, testfilename=None, pardict=None,
            quiet=False, liveoutput=None, chdirpath=None):
        """
        Runs a GSLIB style program using the subprocess module and prints the
        output. On an error, the output is printed and an exception is raised

        The only required parameters are `program` and `parstr`, all other
        parameters are optional.

        Parameters:
            parstr (str): parameters, taken from self if None
            parfile (str or callable): name of parameter file to create, or a callable function
                that returns a unique path/parfile.par string to use for this parfile.
            program (str): name of GSLIB/CCG program to run, taken from self if None
            nogetarg: uses a pipe + communicate for the call instead of arguments
            filehandle: handle for file to write program output to
            logfile (str): filename for a log file to write program output to. If file already
                exists it will overwrite the file. If filehandle is passed then logfile will
                be ignored.
            testfilename (list): name(s) to check for availability prior to
                executing the program
            quiet: if quiet, don't print to let the user know it is calling
            liveoutput (bool): live update the ipython notebook or calling
                script with the output of the called program
            chdirpath (str): Some programs have files they use which prevent
                running multiple instances of the same program from the same
                directory. This can be used to chang the path the program is
                called from to prevent conflicts like this (for example, with kt3d_lva)

        Examples:
            An example for setting up a few key words and running histplt

            >>> histpltpar = '''                  Parameters for HISTPLT
            ...           **********************
            ...
            ... START OF PARAMETERS:
            ... {datafl}          -file with data
            ... {varcol}   0                        -   columns for variable and weight
            ... -1.0     1.0e21              -   trimming limits
            ... {outfl}                   -file for PostScript output
            ...  0.0      -20.0               -attribute minimum and maximum
            ... -1.0                         -frequency maximum (<0 for automatic)
            ... 20                           -number of classes
            ... 0                            -0=arithmetic, 1=log scaling
            ... 0                            -0=frequency,  1=cumulative histogram
            ... 0                            -   number of cum. quantiles (<0 for all)
            ... 3                            -number of decimal places (<0 for auto.)
            ... {varname}                                                    -title
            ... 1.5                          -positioning of stats (L to R: -1 to 1)
            ... -1.1e21                      -reference value for box plot
            ... '''
            >>>
            >>> histplt = gs.Program(program='histplt', parfile='histplt.par')
            >>>
            >>> histplt.run(parstr=histpltpar.format(datafl=datafl.flname,
            ...                                      varcol=datafl.gscol('Bitumen'),
            ...                                      varname='Bitumen',
            ...                                      outfl='histplt_bitumen.ps'))

        """
        if program is None:
            program = self.program
        else:
            program = self.__check_program(program)
        if parstr is None:
            parstr = self.parstr
        if nogetarg is None:
            nogetarg = self.nogetarg
        if parfile is None:
            parfile = self.parfile
        # Update the working dictionary
        workingdict = self.defaultdict.copy()
        if pardict is not None:
            workingdict.update(pardict)

        # Logic Checks and figureing out defaults
        if liveoutput is None and filehandle is None and logfile is None:
            liveoutput = True
        elif liveoutput is None and logfile:
            liveoutput = True
        elif liveoutput is None and filehandle:
            liveoutput = False
        if filehandle:
            if logfile:
                warnings.warn('filehandle was passed so logfile is being ignored')
                logfile = None

        if isinstance(logfile, str):
            consoleout = open(logfile, 'w')
        elif logfile:
            raise TypeError('logfile must be string')

        # parfile could be a callable that generates tempfiles just before the program is run
        # e.g. in parallel
        try:
            parfile = parfile()
            shouldclean = True
        except TypeError:
            shouldclean = False
            pass

        program_call = [program, parfile]
        if not quiet:
            print('Calling: ', program_call)

        # remove leading whitespace on each line of parstr so it can be nested correctly in a
        # function or code-folded in the notebook!
        parstr = dedent_parstr(parstr)

        # Format the parstr first to catch KerError for printing if need be
        try:
            parstr = parstr.format(**workingdict)
        except KeyError as err:
            errormsg = 'The following Dictionary Key was not found: ' + str(err)
            raise KeyError(errormsg)

        # Create the parameter file
        try:
            if chdirpath is None:
                try:
                    with open(parfile, 'w', newline='', encoding='utf8') as pf:
                        pf.write(parstr)
                except:
                    with open(parfile, 'w') as pf:
                        pf.write(parstr)
            else:
                try:
                    with open(chdirpath + '/' + parfile, 'w', newline='', encoding='utf8') as pf:
                        pf.write(parstr)
                except:
                    with open(chdirpath + '/' + parfile, 'w') as pf:
                        pf.write(parstr)
        except:
            errormsg = 'Could not create parameter file: ' + parfile
            raise Exception(errormsg)
        if testfilename is not None:
            time.sleep(random.random())
            # Test if this file is open and available up to a max of 5 times
            # before just running anyway
            nattempts = 0
            while nattempts < 5:
                if not isinstance(testfilename, str):
                    try:
                        with open(testfilename, 'r'):
                            testfilename.readline()
                        break
                    except:
                        nattempts += 1
                        time.sleep(1.0 + random.random() * nattempts)
                else:
                    try:
                        for testfile in testfilename:
                            with open(testfile, 'r'):
                                testfile.readline()
                        break
                    except:
                        nattempts += 1
                        time.sleep(1.0 + random.random() * nattempts)

        if nogetarg:
            process = subprocess.Popen([program], stdout=subprocess.PIPE, stdin=subprocess.PIPE,
                                       stderr=subprocess.PIPE)
            # message = process.communicate(input=bytes(parfile+'\n', 'ascii'))
            try:  # py3.x
                message = process.communicate(input=bytes(parfile + ' \n', 'ascii'))
            except:
                try:  # py2.7
                    message = process.communicate(input=unicode(parfile + ' \n', 'ascii'))
                except:
                    raise Exception('Something went wrong with calling ' + str(program))
            # print(message[0].decode('ascii'))
            if filehandle is None:
                if logfile is None:
                    print(''.join([msg.decode('ascii') for msg in message]))
                else:
                    consoleout.write(''.join([msg.decode('ascii') for msg in message]))
                    if liveoutput:
                        print(''.join([msg.decode('ascii') for msg in message]))
            else:
                filehandle.write(''.join([msg.decode('ascii') for msg in message]))
                if liveoutput:
                    print(''.join([msg.decode('ascii') for msg in message]))

        else:
            # The standandard for single-threaded calls to executables, allows for live output
            # and catches common errors
            try:
                p = subprocess.Popen(program_call, cwd=chdirpath,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.STDOUT, bufsize=1,
                                     universal_newlines=True)
                fulloutput = ''
                for line in p.stdout:    # still not working correctly output not
                    if liveoutput:
                        print(line, end='')  # constantly fed to the ipython notebook..
                    if "forrtl: The pipe has been ended." in line:
                        print("**************************************************************\n"
                              "WARNING: it looks like the program %s does not use `getarg`\n"
                              "Consider the argument `nogetarg=True` for the call to %s.run()!\n"
                              "**************************************************************\n" %
                              (self.program, self.program))
                        if shouldclean:
                            os.remove(parfile)
                        return
                    fulloutput += line
                if filehandle:
                    filehandle.write(fulloutput)
                elif logfile:
                    consoleout.write(fulloutput)
            except subprocess.CalledProcessError:
                # If the program crashed, raise an exception
                errormsg = 'ERROR executing: ' + str(program_call)
                if logfile:
                    consoleout.write(errormsg)
                    consoleout.close()
                raise Exception(errormsg)
            except FileNotFoundError:
                # Notify that perhaps this executable doesn't exist in the path ...
                errormsg = 'ERROR calling: ' + str(program_call[0]) + '. Check path to exe'
                if logfile:
                    consoleout.write(errormsg)
                    consoleout.close()
                raise Exception(errormsg)
            except:
                # This subprocess.check_output is required to call exe's in parallel with
                # gs.runparallel()
                # filehandles does not work in parallel
                try:
                    # Call the subprocess and buffer the output
                    subprocess_output = subprocess.check_output(program_call, cwd=chdirpath)
                except subprocess.CalledProcessError:
                    # If the program crashed, raise an exception
                    errormsg = 'ERROR executing: ' + str(program_call)
                    if logfile:
                        consoleout.write(errormsg)
                        consoleout.close()
                    raise Exception(errormsg)
                except FileNotFoundError:
                    # Notify that perhaps this executable doesnt exist in the path ...
                    errormsg = 'ERROR calling: ' + str(program_call[0]) + '. Check path to exe'
                    if logfile:
                        consoleout.write(errormsg)
                        consoleout.close()
                    raise Exception(errormsg)
                # Print out the decoded output from the subprocess if successful
                try:
                    decoded_output = subprocess_output.decode("utf-8")
                except UnicodeDecodeError:
                    print('Warning: UnicodeDecodeError decoding subprocess output with utf-8.')
                    decoded_output = ''
                if filehandle:
                    filehandle.write(decoded_output)
                if logfile:
                    consoleout.write(decoded_output)
                else:
                    print(decoded_output)
                # Check for signal of GSLIB crash in output in case the program
                # doesn't return a non-zero error code
                if 'FAILED' in decoded_output:
                    print('FAILED found in gslib output - logging.')
                    errormsg = 'ERROR executing: ' + ' '.join(program_call)
                    if logfile:
                        consoleout.write(errormsg)
                        consoleout.close()
                    raise Exception(errormsg)
        # Close logfile if it is open
        if logfile and not consoleout.closed:
            consoleout.close()
        if shouldclean:
            os.remove(parfile)

    def writepar(self, parstr=None, parfile=None, pardict=None):
        '''
        Writes out the parameter file without running the program, which can be
        helpful for checking

        Parameters:
            parstr (str): This is the parameter file string initiated with the program.
            parfile (str): file name or path to save the parameter file to.
            pardict (dict): Dictionary for the variables in the parameter file.
        '''
        # Keep the default parameter string if no new string provided
        if parstr is None:
            parstr = self.parstr
        if parfile is None:
            parfile = self.parfile
        # Update the working dictionary
        workingdict = self.defaultdict.copy()
        if pardict is not None:
            workingdict.update(pardict)
        # Create the parameter file
        try:
            with open(parfile, 'w', newline='', encoding='utf8') as pf:
                pf.write(parstr.format(**workingdict))
        except:
            errormsg = 'Could not create parameter file: ' + parfile
            raise Exception(errormsg)

    def getparfile(self, quiet=True):
        """
        Get the parfile from this program by copying to the clipboard or
        printing the parfile. This replaces the need to pre-execute a program
        to get the parfile, but relies on CCG programs being properly configured
        to generate the correct parfile upon first execution.

        This function requires `pyperclip`. This is a dependency of pygeostat, but
        can be installed with::

            > pip install pyperclip

        Parameters:
            quiet (bool) : The function will copy the parameter file to the
                clipboard with the block quotes as default (quiet = True)

        """
        import tempfile
        with tempfile.TemporaryDirectory(dir=os.getcwd()) as dirname:
            # Run program to automatically generate a parfile
            try:
                p = subprocess.Popen(self.program, stdout=subprocess.PIPE,
                                     stdin=subprocess.PIPE, stderr=subprocess.STDOUT,
                                     cwd=dirname)
                try:  # py3.x
                    message = p.communicate(input=bytes(' \n', 'ascii'))
                except:
                    try:
                        message = p.communicate(input=u' \n')
                    except:
                        raise Exception('Something went wrong with calling ' + self.program)
            except FileNotFoundError:
                raise FileNotFoundError('The program ' + self.program + ' was not found, '
                                        'check definitions!')
            # assume that the parfile is called "self.program + '.par' "
            if os.path.isfile(dirname + '/' + self.program + '.par'):
                parfl = dirname + '/' + self.program + '.par'
            else:  # grab the newest .par file that was created in this directory
                import glob
                try:
                    parfl = max(glob.iglob(dirname + '/*.par'), key=os.path.getctime)
                except ValueError:
                    raise ValueError('%s did not create a parfile! You may require the parfile '
                                     'from the ccg knowledgebase. Try: '
                                     'https://ccgsrv.geostats.ualberta.ca/ccgkb/doku.php?'
                                     'do=search&id=%s' % (self.program, self.program))
            # read the par file
            with open(parfl, 'r') as fl:
                parstr = fl.read()
            os.remove(parfl)

        # the panda's copy stopped working... no idea why. pyperclip seems to do a
        # better job anyway!
        if quiet:
            try:
                import pyperclip
                progstr = self.program
                if '/' in progstr:
                    progstr = progstr[progstr.rfind('/') + 1:]
                if '.' in progstr:
                    progstr = progstr[:progstr.rfind('.')]
                pyperclip.copy('parstr = """' + parstr + '"""')
                print(parfl + ' has been copied to the clipboard')
            except (ModuleNotFoundError, ImportError):
                print('Looks like pyperclip is not installed, use:\n',
                      '>>> pip install pyperclip\n',
                      'to install. In the meantime, here is the par file!\n')
                print(parstr)
        else:
            print(parstr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""A collection of utility functions for variogram plotting"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
import pandas as pd
import numpy as np


def get_uniquevarids(data, mode='print', source=None):
    """
    Enumerate the unique variograms within a output gslib variogram files based on the number of
    directions and variables found. Prints the unique variograms found to the screen so a varid can
    be determined for :func:`gs.varplt() <pygeostat.plotting.varplt.varplt>` or
    :func:`gs.varpltsim() <pygeostat.plotting.varpltsim.varpltsim>`. Alternatively, a dictionary
    containing the Variogram IDs and their corresponding 'Variogram Index' values and a list
    containing the directions found can be returned.

    Files from varcalc, varmodel, and varsim are accepted.

    .. note:
        This function relies on the arrangement of data within the output files to determine some
        parameters. If it is having some difficulties, the parameter ``source`` can be used to
        possibly solve any issues.

    Parameters:
        data (pd.DataFrame):  Dataframe containing the output data from gslib variogram programs.
            The columns 'Variogram Index', 'Variogram Number', 'Calculation Azimuth', and
            'Calculation Dip' must be unedited; specifically their headers.
        mode (str): Specify the required output from the function. If ``'print'`` is passed,
            the enumerated variogram IDs will be printed with their corresponding variable and
            direction information. If ``'ref'`` is passed, a dictionary is build for each variogram
            ID containing their corresponding 'Variogram Index' values and a direction
            list``dirs``. source (str): Manually set what the data source is. Permissible input
            is: ``'varcalc'``, ``'varmodel'``, and ``'varsim'``.

    Returns:
        varids (dict): Returned if ``mode`` is set to ``'ref'``. A dictionary is build containing
        each variogram ID and their corresponding 'Variogram Index' value or values if multiple
        realizations are found.

    Returns:
        dirs (list): Returned if ``mode`` is set to ``'ref'``. A list of directions found within
        the variogram data

    Examples:
        Load variogram output into python:

        >>> varsimdat = gs.DataFile('varsim_reals.out', readfl=True)

        Check the variogram IDs that will be used by varpltsim:

        >>> gs.get_uniquevarids(varsimdat.data)
            Variogram ID: 1 ... Variable: 1, Azimuth 90, Dip 0
            Variogram ID: 2 ... Variable: 2, Azimuth 90, Dip 0
            Variogram ID: 3 ... Variable: 3, Azimuth 90, Dip 0

        Return the varids and print the index values attached to each Variogram ID:

        >>> varids, dirs = gs.get_uniquevarids(varsimdat.data, mode='ref')
        >>> print(varids[1])
        [1, 4, 7, ..., 292, 295, 298]

        The list ``varids`` can now be used to repeadidly plot :func:`gs.varplt()
        <pygeostat.plotting.varplt.varplt>`, which is what :func:`gs.varpltsim()
        <pygeostat.plotting.varpltsim.varpltsim>` does.

        Have a look at the list ``dirs`` which was generated above:

        >>> print(dirs)
        [(90, 0)]

        The output ``dirs`` can be useful in situations such as generating plot titles.

    """
    # Infer and extract the required columns
    if 'Variogram Index' in data.columns:
        vario_index = data['Variogram Index']
    else:
        print("Error: The column 'Variogram Index' could not be found")
        return
    if 'Variogram Number' in data.columns:
        vario_num = data['Variogram Number']
    else:
        print("Error: The column 'Variogram Number' could not be found")
        return
    if 'Calculation Azimuth' not in data.columns and 'Calculation Dip' not in data.columns:
        print("Error: One or more of the columns 'Calculation Azimuth' and 'Calculation Dip' could"
              "not be found")
        return
    # Get the number of directions
    dirs = []
    for direction, _ in data.groupby(['Calculation Azimuth', 'Calculation Dip'], sort=False):
        dirs.append(direction)
    ndir = len(dirs)
    # Get the number of lags per variogram
    nlegs = len(vario_index[vario_index == 1])
    # Get the number of variables
    nvarios = len(vario_index.unique())
    nvar = len(vario_num.unique()) // ndir
    if source is 'varmodel' or nvarios == 1 or nvarios == ndir:
        nvar = 1
        nreal = 1
    elif source is 'varsim':
        nreal = vario_num[vario_num == 1].count() // nlegs
    else:
        nreal = vario_num[vario_num == 1].count() // nlegs // ndir
    # Check that the number of variograms in the file match what should be there.
    calcnvarios = nreal * nvar * ndir
    if calcnvarios != nvarios:
        print("Error: The number of directions, variables, and realizations do not match the",
              "number of\n       variograms found in the file")
        return
    # Build a reference list of variogram IDs and the 'Variogram Index' values that correspond to
    # that variogram ID
    varids = {}
    i = 1
    for j in range(1, nvar + 1):
        for k in range(0, ndir):
            varids[i] = list(range(i, nvarios + 1)[::(nvar * ndir)])
            if mode is 'print':
                print("Variogram ID: %s ... Variable: %s, Azimuth %s, Dip %s"
                      % (i, j, dirs[k][0], dirs[k][1]))
            i += 1
    # Return some variables if in the right mode
    if mode is 'ref':
        return varids, dirs


def trimylim(simdat, ylim):
    """
    Remove realization experimental variogram points beyond a variogram value so they won't be
    plotted.

    Parameters:
        simdat (pd.DataFrame):
        ylim (float): Value to trim points after
    """
    # Get the column containing the index
    if 'Variogram Index' in simdat.columns:
        index_name = 'Variogram Index'
    elif 'Index' in simdat.columns:
        index_name = 'Index'
    trimdata = []
    for index in np.unique(simdat[index_name]):
        # Figure out where to start trimming
        data = simdat[simdat[index_name] == index]
        indices = data[data.rolling(center=False, window=5).mean()['Variogram Value'] > ylim]
        indices = indices.index.tolist()
        # Trim if required
        if len(indices) > 0:
            minix = min(indices)
            # Dump the rows after the ceiling is reached
            trimdata.append(data.iloc[:minix, :])
        else:
            # Dump the data even if it doesn't need trimming
            trimdata.append(data)
    # Rebuild the trimmed DataFrame
    simdat = pd.concat(trimdata)

    return simdat



'''
The :mod:`pygslib.data` module includes DataFile functions.
'''
from __future__ import absolute_import, division, print_function

from .data import *
from .iotools import *
from .grid_definition import *
from .h5_io import *



# !/usr/bin/env python
#  -*- coding: utf-8 -*-
"""accuracy_plot plots the fraction of the true values that fall within a specified probability interval"""

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
from . set_style import set_plot_style

@set_plot_style
def accuracy_plot(truth=None, reals=None, mik_thresholds=None, acctype='sim', probability_increment=0.05,
		   figsize=None, title=None, xlabel=None, ylabel=None, stat_blk='standard',
		   stat_xy=(.95, .05), stat_fontsize=None, ms=5, grid=None, axis_xy=None, ax=None,
		   plot_style=None, custom_style=None, output_file=None, **kwargs):
	"""
	Accuracy plot based on probability intervals quantified using an estimation technique e.g kriging.
	Currently, this plotting tool works based on using realizations to quantify the distrbution of estimation.


	Two statistics block sets are available: ``'minimal'`` and the default ``'standard'``. The
	statistics block can be customized to a user defined list and order. Available statistics are
	as follows:

	>>> ['ndat', 'nint', 'avgvar', 'mse', 'acc', 'pre', 'goo']

	Please review the documentation of the :func:`gs.PlotStyle.set_style()
	<pygeostat.pygeostat_parameters.PlotStyle.set_style>` and :func:`gs.export_image()
	<pygeostat.plotting.export_image.export_image>` functions for details on their parameters so that
	their use in this function can be understood.

	Keyword Arguments:
	   
		truth: Tidy (long-form) 1D data where a single column containing the true values.
			A pandas dataframe/series or numpy array can be passed
		reals: Tidy (long-form) 2D data where a single column contains values from a single
			realizations and each row contains the simulated values from a single truth location.
			A pandas dataframe or numpy matrix can be passed
		mik_thresholds (np.ndarray): 1D array of the z-vals ``mik_thresholds`` corresponding to the
			probabilities defined in reals for each location
		acctype (str): Currently ``sim`` and ``mik`` are valid. if ``mik_thresholds`` is passed the
			type is assumed to be ``mik``
		probability_increment (float): Probability increment used during accuracy_plot calculation
		figsize (tuple): Figure size (width, height)
		title (str): Title for the plot
		xlabel (str): X-axis label
		ylabel (str): Y-axis label
		stat_blk (bool): Indicate if statistics are plotted or not
		stat_xy (float tuple): X, Y coordinates of the annotated statistics in figure space. The
			coordinates specify the top right corner of the text block
		stat_fontsize (float): the fontsize for the statistics block. If None, based on
			gs.Parameters['plotting.stat_fontsize']. If less than 1, it is the fraction of the
			matplotlib.rcParams['font.size']. If greater than 1, it the absolute font size.
		ms (float): Size of scatter plot markers
		grid(bool): plots the major grid lines if True. Based on gs.Parameters['plotting.grid']
			if None.
		axis_xy (bool): converts the axis to GSLIB-style axis visibility (only left and bottom
			visible) if axis_xy is True. Based on gs.Parameters['plotting.axis_xy'] if None.
		ax (mpl.axis): Matplotlib axis to plot the figure
		pltstyle (str): Use a predefined set of matplotlib plotting parameters as specified by
			:class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
			to turn it off
		cust_style (dict): Alter some of the predefined parameters in the ``pltstyle`` selected.
		output_file (str): Output figure file name and location
		**kwargs: Optional permissible keyword arguments to pass to
			:func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`

	Returns:
		ax (ax): Matplotlib Axes object with the cross validation plot

	Examples:


	A simple call using truth and realization data:

	.. plot::

		import pygeostat as gs
		data_file = gs.ExampleData('accuracy_plot')
		reals = data_file[list(data_file.columns[1:])].values
		truth = data_file[list(data_file.columns)[0]].values
		gs.accuracy_plot(truth=truth, reals=reals)

	"""
	import numpy as np
	from ..statistics.utils import accsim, accmik
	from .utils import setup_plot, get_statblk, format_plot, _set_stat_fontsize
	from .export_image import export_image
	import matplotlib as mpl
	from .. pygeostat_parameters import Parameters

	# Sanity checks based on the provided input arguments
	simcheck = [(val is not None) for val in [truth, reals]]
	mikcheck = [(val is not None) for val in [mik_thresholds]]
	if any(simcheck) and not all(simcheck):
		raise ValueError("The arguments `truth` and `reals` are required")
	if any(mikcheck) and not all(simcheck):
		raise ValueError("Must have `truth`, `reals` and `mik_thresholds` for MIK !")
	if all(simcheck) and acctype.lower() != 'sim':
		acctype = 'sim'
	if all(mikcheck) and acctype == 'sim':
		acctype = 'mik'
	# Handle accuracy_plot-sim functionality
	if acctype.lower() == 'sim':
		propavg, statlist = accsim(truth, reals, probability_increment)
		x = propavg['ProbInt'].values
		y = propavg['FracIn'].values
	elif acctype.lower() == 'mik':
		propavg, statlist = accmik(truth, mik_thresholds, reals, probability_increment)
		x = propavg['ProbInt'].values
		y = propavg['FracIn'].values
	else:
		raise ValueError ('Wrong acctype was provided.')

	fig, ax, cax = setup_plot(ax, cbar=False, figsize=figsize)
	ax.set_aspect('equal')
	# Plot the figure
	for point in zip(x, y):
		ax.plot((point[0], point[0]), (point[1], point[0]), color='red', zorder=0, lw=0.25)
	ax.scatter(x, y, c='k', s=ms)
	ax.plot((1, 0), (1, 0), zorder=0, lw=0.5, color='k')
	# Configure plot
	ax.axis('equal')
	if xlabel is None:
		xlabel = 'Probability Interval'
	if ylabel is None:
		ylabel = 'Fraction in Interval'
	format_plot(ax, xlabel=xlabel, ylabel=ylabel, title=title,
			  grid=grid, xlim=(0, 1.0), ylim=(0, 1.0))
	# Plot stats
	if stat_blk:
		statlist = {'ndat': '$n = {}$'.format(len(truth)),
					'nint': '$n_p = {}$'.format(len(x)),
					'avgvar': '$U = {:0.3e}$'.format(statlist['avgvar']),
					'mse': '$MSE = {:0.3e}$'.format(statlist['mse']),
					'acc': '$Accuracy = {:0.3e}$'.format(statlist['acc']),
					'pre': '$Precision = {:0.3e}$'.format(statlist['pre']),
					'goo': '$Goodness = {:0.3e}$'.format(statlist['goo'])}
		statsets = {'minimal': ['nint'],
					'standard': ['nint', 'avgvar', 'mse'],
					"all": ['ndat', 'nint', 'avgvar', 'mse', 'acc', 'pre', 'goo']}
		if len(truth) > 0:
			statsets['standard'].insert(0, 'ndat')
		if isinstance(stat_blk, str):
			stat_blk = [stat_blk]
		# Get the stat block text
		txtstats, stat_xy, ha, va = get_statblk(stat_blk, statsets, statlist, stat_xy)
		# Write the stat block to the plot
		stat_fontsize = _set_stat_fontsize(stat_fontsize)
		ax.text(stat_xy[0], stat_xy[1], txtstats, va=va, ha=ha, transform=ax.transAxes,
				fontsize=mpl.rcParams['font.size'] - 0.5)
	# Export figure
	if output_file or ('pdfpages' in kwargs):
		export_image(output_file, **kwargs)

	return ax


#!/usr/bin/env python

# -*- coding: utf-8 -*-
"""Location map plotting routine using matplotlib"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
from . set_style import set_plot_style
from .. pygeostat_parameters import Parameters

@set_plot_style
def location_plot(data, x=None, y=None, z=None, var=None, dhid = None, catdata=None, allcats=True, cbar=True,
		   cbar_label=None, catdict=None, cmap=None, cax=None, vlim=None, title=None,  plot_collar = True, collar_marker='x',
		   collar_offset = 0, lw = None, xlabel=None, ylabel=None, unit=None, griddef=None, slice_number=None, orient='xy',
		   slicetol=None, xlim=None, ylim=None, ax=None, figsize=None, s=None, marker='o',
		   rotateticks=None, sigfigs=3, grid=None, axis_xy=None, aspect=None, plot_style=None,
		   custom_style=None, output_file=None, out_kws=None, return_cbar=False, return_plot=False,
		   **kwargs):
	"""
	location_plot displays scattered data on a 2-D XY plot. To plot gridded data with or without
	scattered data, please see :func:`gs.slice_plot()<pygeostat.plotting.slice_plot>`.

	The only required parameter is ``data`` if it is a
	:class:`gs.DataFile <pygeostat.data.data.DataFile>` that contains the necessary coordinate
	column headers, data, and if required, a pointer to a valid
	:class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>` class. All other parameters are
	optional. If ``data`` is a :class:`gs.DataFile <pygeostat.data.data.DataFile>` class and does
	not contain all the required parameters or if it is a long-form table, the following
	parameters will need to be pass are needed: ``x``, ``y``, ``z``, and ``griddef``. The three
	coordinate parameters may not be needed depending on what ``orient`` is set to and of course
	if the dataset is 2-D or 3-D. The parameter ``griddef`` is required if ``slicetol`` or
	`` slice_number`` is used. If parameter ``slice_number`` and ``slicetol`` is not set then the default
	slice tolerance is half the cell width. If a negative ``slicetol`` is passed or slice_number is set
	to None then all data will be plotted. ``slicetol`` is based on coordinate units.

	The values used to bound the data (i.e., vmin and vmax) are automatically calculated by default.
	These values are determined based on the number of significant figures and the sliced data;
	depending on data and the precision specified, scientific notation may be used for the colorbar
	tick lables. When point data shares the same colormap as the gridded data, the points displayed
	are integrated into the above calculation.

	Please review the documentation of the :func:`gs.set_style()
	<pygeostat.plotting.set_style.set_style>` and :func:`gs.export_image()
	<pygeostat.plotting.export_image.export_image>` functions for details on their parameters so that
	their use in this function can be understood.

	Parameters:
		data (pd.DataFrame or gs.DataFile): data containing coordinates and (optionally) var
		x (str): Column header of x-coordinate. Required if the conditions discussed above are not
			met
		y (str): Column header of y-coordinate. Required if the conditions discussed above are not
			met
		z (str): Column header of z-coordinate. Required if the conditions discussed above are not
			met
		var (str): Column header of the variable to use to colormap the points. Can also be a list
			of or single permissible matplotlib colour(s). If None and data is a DataFile,
			based on DataFile.variables if len(DataFile.variables) == 1. Otherwise, based on
			Parameters['plotting.location_plot.c']
		dhid (str): Column header of drill hole ID.
		catdata (bool): Force categorical data
		catdict (dict): Dictionary containing the enumerated IDs alphabetic equivalent, which is
			drawn from Parameters['data.catdict'] if None
		allcats (bool): ensures that if categorical data is being plotted and plotted on slices,
			that the categories will be the same color between slices if not all categories are
			present on each slice
		cbar (bool): Indicate if a colorbar should be plotted or not
		cbar_label (str): Colorbar title
		cmap (str): A matplotlib colormap object or a registered matplotlib or pygeostat colormap
			name.
		cax(Matplotlib.ImageGrid.cbar_axes): color axis, if a previously created one should be used
		vlim (float tuple): Data minimum and maximum values
		title (str): Title for the plot. If left to it's default value of ``None`` or is set to
			``True``, a logical default title will be generated for 3-D data. Set to ``False`` if
			no title is desired.
		plot_collar (bool): Option to plot the collar if the orient is xz or yz and the dhid is provided/inferred
		collar_marker (str): One of the permissible matplotlib markers, like 'o', or '+'... and others.
		lw (float): Line width value if the orient is xz or yz and the dhid is provided/inferred. Because lines or plotted instead of points. 
		xlabel (str): X-axis label
		ylabel (str): Y-axis label
		unit (str): Unit to place inside the axis label parentheses
		griddef (GridDef): A pygeostat GridDef class created using
			:class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Required if using the
			argument ``slicetol``
		orient (str): Orientation to slice data. ``'xy'``, ``'xz'``, ``'yz'`` are t he only accepted
			values
		slice_number (int): Grid cell location along the axis not plotted to take the slice of data to
			plot. None will plot all data
		slicetol (float): Slice tolerance to plot point data (i.e. plot +/- ``slicetol`` from the
			center of the slice). Any negative value plots all data. Requires ``slice_number``. If a
			``slice_number`` is passed and no ``slicetol`` is set, then the default will half the cell
			width based on the griddef.
		xlim (float tuple): X-axis limits. If None, based on data.griddef.extents(). If
			data.griddef is None, based on the limits of the data.
		ylim (float tuple): Y-axis limits. If None, based on data.griddef.extents(). If
			data.griddef is None, based on the limits of the data.
		ax (mpl.axis): Matplotlib axis to plot the figure
		figsize (tuple): Figure size (width, height)
		s (float): Size of location map markers
		marker (str): One of the permissible matplotlib markers, like 'o', or '+'... and others.
		rotateticks (bool tuple): Indicate if the axis tick labels should be rotated (x, y)
		sigfigs (int): Number of sigfigs to consider for the colorbar
		grid (bool): Plots the major grid lines if True. Based on Parameters['plotting.grid']
			if None.
		axis_xy (bool): converts the axis to GSLIB-style axis visibility (only left and bottom
			visible) if axis_xy is True. Based on Parameters['plotting.axis_xy'] if None.
		aspect (str): Set a permissible aspect ratio of the image to pass to matplotlib. If None,
			it will be 'equal' if each axis is within 1/5 of the length of the other. Otherwise,
			it will be 'auto'.
		plot_style (str): Use a predefined set of matplotlib plotting parameters as specified by
			:class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
			to turn it off
		custom_style (dict): Alter some of the predefined parameters in the ``plot_style`` selected.
		output_file (str): Output figure file name and location
		out_kws (dict): Optional dictionary of permissible keyword arguments to pass to
			:func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`
		return_cbar (bool): Indicate if the colorbar axis should be returned. A tuple is returned
			with the first item being the axis object and the second the cbar object.
		return_plot (bool): Indicate if the plot from scatter should be returned. It can be used
							to create the colorbars required for subplotting with the ImageGrid()
		**kwargs: Optional permissible keyword arguments to pass to matplotlib's scatter function

	Returns:
		ax (ax): Matplotlib axis instance which contains the gridded figure

	Returns:
		cbar (cbar): Optional, default False. Matplotlib colorbar object

	**Examples:**

	A simple call:

	.. plot::

		import pygeostat as gs
		data_file = gs.ExampleData('point3d_ind_mv')
		gs.location_plot(data_file)

	|

	A simple call using a variable to color the data:

	.. plot::

		import pygeostat as gs
		data_file = gs.ExampleData('point3d_ind_mv')
		gs.location_plot(data_file, var = 'Phi')

	|

	Plotting along xz/yz orientation and use line plots based on drill hole id

	.. plot::

		import pygeostat as gs
		# load data
		data_file = gs.ExampleData('point3d_ind_mv')
		gs.location_plot(data_file, var='Phi', orient='yz', aspect =5, plot_collar = True)
		
	|

	Location plot for a categorical variable

	.. plot::

		import pygeostat as gs
		# load data
		data_file = gs.ExampleData('point3d_ind_mv')
		gs.location_plot(data_file, var='Lithofacies', orient='yz', aspect =5, plot_collar = True)

	|
	
	"""
	from . utils import (_spatial_labels, _spatial_pointdata, _spatial_slice, _format_tick_labels, setup_plot,
						 _spatial_orient2fig, format_plot, _get_cmap, _spatial_aspect, get_contcbarargs)

	from .. pygeostat_parameters import Parameters
	from . export_image import export_image
	from ..data.data import DataFile
	from .cmaps import avail_palettes
	from matplotlib.collections import LineCollection
	
	# Handle dictionary defaults
	if not out_kws:
		out_kws = dict()

	# Infer dhid column if not provided
	drill_plot_format = False
	if dhid is None:
		if isinstance(data, DataFile):
			dhid = data.dh
	else:
		if not dhid in data.columns:
			raise ValueError('The provided drill hole id column, {} does not exist'.format(dhid))

	# Handle var
	if var is None:
		if isinstance(data, DataFile):

			if data.variables is not None:
				if isinstance(data.variables, str):
					var = data.variables
			elif isinstance(data.cat, str):
				var = data.cat
		if var is None:
			var = Parameters['plotting.location_plot.c']
	
	# Handle title
	if title is None:
		if var in data.columns:
			title = var
	# Determine the x, y, z and griddef with error checking
	data, x, y, z, griddef = _spatial_pointdata(data, orient, x, y, z, griddef)

	# Determine the data to plot based on grid def and slice info
	if dhid is None:
		pointx, pointy, pointvar = _spatial_slice(data, var, x, y, z, griddef, orient, slice_number, slicetol)
	else:
		pointx, pointy, pointvar, point_dh = _spatial_slice(data, var, x, y, z, griddef, orient, slice_number, slicetol, dhid)

	if pointx is None:
		if griddef is None:
			raise ValueError('At least two coordinates are required for the location_plot')
		else:
			print('Note: There is no data point detected within the provided slice number and/or slice tolerance')

	# Setup the color
	if pointvar is None:
		try:
			pointvar = mpl.colors.ColorConverter().to_rgb(var)
			if int(mpl.__version__.split('.')[0]) >= 3:
				pointvar = [pointvar]
			cbar = False
		except Exception:
			if griddef is None:
				raise ValueError(
					'var is not a variable in the passed data and not a valid Matplotlib color {}'.format(var))
			else:
				pointvar = [(0.0, 0.0, 0.0)] # Use a default color since there is no pint within the grid
				cbar = False

	# Get default colormaps or palettes. If there are Parameters['plotting.assumecat'] or less
	# unique data, assume categorical
	if isinstance(var, str):
		if var in data.columns:
			if allcats:
				alldata = data[var]
			else:
				alldata = pointvar
			ncat = len(np.unique(alldata[np.isfinite(alldata)]))
		else:
			ncat = None
	else:
		ncat = None
		cmap = None
	if catdata is not True:
		if cmap is None and catdata is None and ncat is not None and \
				ncat < Parameters['plotting.assumecat']:
			catdata = True
		# Get palette from pygeostat
		elif cmap in avail_palettes:
			catdata = True
		elif ncat is not None and ncat <= Parameters['plotting.assumecat'] and catdata is None:
			catdata = True
		else:
			catdata = False
	if cmap is None:
		cmap = _get_cmap(cmap, catdata, ncat)
	# Set-up categorical parameter if required
	if catdata:
		ticklabels = np.unique(alldata[np.isfinite(alldata)]).astype(int)
		if catdict is None:
			catdict = Parameters['data.catdict']
		if isinstance(catdict, dict):
			if len(ticklabels) != len(catdict):
				# modify the cmap to store the colors as if cmap is generated from all cats
				cmap = _get_cmap(cmap, True, len(catdict))
				cmap = mpl.colors.ListedColormap(
					[clr for cat, clr in zip(catdict, cmap.colors) if cat in ticklabels]
				)
		vlim = (0, ncat)
		ticklocs = np.arange(ncat) + 0.5
		if cmap:
			if isinstance(cmap, str):
				cmap = _get_cmap(cmap, catdata, ncat)
			dump = pointvar.copy()
			for i in range(ncat):
				dump[pointvar == ticklabels[i]] = i
			pointvar = dump
		if isinstance(catdict, dict):
			ticklabels = [catdict[cat] for cat in ticklabels]
	# Set-up some parameters
	if not catdata and cmap is not False and cbar:
		vlim, ticklocs, ticklabels = get_contcbarargs(pointvar, sigfigs, vlim)
	if vlim is None:
		vlim = (None, None)
	# Set-up plot if no axis is supplied using the ImageGrid method if required or the regular way
	fig, ax, cax = setup_plot(ax, cax=cax, cbar=cbar, figsize=figsize)
	if s is None:
		s = Parameters['plotting.location_plot.s']

	# used for plot extents and get the two coordinates being plotted
	figx, figy, _ = _spatial_orient2fig(orient, x, y, z)

	if orient.lower() == 'xy' or not dhid:
		plot = ax.scatter(pointx, pointy, c=pointvar, cmap=cmap, vmin=vlim[0],
						vmax=vlim[1], s=s, marker=marker, **kwargs)
	else: # Plot lines if the dhid was inferred/provided and the orientation is xz or yz
		drill_plot_format = True
		data_temp = pd.DataFrame({figx:pointx, figy: pointy, var:pointvar, dhid: point_dh})
		grouped = data_temp.groupby(dhid, sort=False)

		if lw is None:
			lw = Parameters['plotting.location_plot.lw']
		for _, group in grouped:
			if plot_collar:
				x_points = group[figx][group[figx].index.min()]
				y_points = group[figy][group[figy].index.min()] + collar_offset
				if var in data.columns:
					collar_var = group[var][group[var].index.min()]
					norm = plt.cm.colors.Normalize(*vlim)
					_cmap = plt.cm.get_cmap(cmap)
					collar_var = [_cmap(norm(collar_var))]
					plot = ax.scatter(x_points, y_points, c=collar_var, cmap=cmap, vmin=vlim[0], vmax=vlim[1], s=s,
							marker=collar_marker, **kwargs)
				else:
					plot = ax.scatter(x_points, y_points, c=pointvar, vmin=vlim[0], vmax=vlim[1], s=s,
							marker=collar_marker, **kwargs)

			points = np.array([group[figx], group[figy]]).T.reshape(-1, 1, 2)
			segments = np.concatenate([points[:-1], points[1:]], axis=1)

			if var in data.columns:
				lc = LineCollection(segments, cmap=cmap, norm=plt.Normalize(vlim[0], vlim[1]))
				lc.set_array(group[var])
			else:
				lc = LineCollection(segments, color=var, norm=plt.Normalize(vlim[0], vlim[1]))
			lc.set_lw(lw)

			ax.add_collection(lc)

	# Plot labels
	ax = _spatial_labels(ax, orient, griddef, slice_number, title, xlabel, ylabel, unit, sigfigs)
	
	
	if xlim is None:
		if griddef:
			if orient in ['xy', 'xz']:
				xlim = griddef.extents()[0]
			else:
				xlim = griddef.extents()[1]
		else:
			pad = (data[figx].max() - data[figx].min()) * 0.025
			xlim = (data[figx].min() - pad, data[figx].max() + pad)
	if ylim is None:
		if griddef:
			if orient == 'xy':
				ylim = griddef.extents()[1]
			else:
				ylim = griddef.extents()[2]
		else:
			pad = (data[figy].max() - data[figy].min()) * 0.025
			ylim = (data[figy].min() - pad, data[figy].max() + pad)
	if aspect is None:
		aspect = _spatial_aspect(xlim, ylim)

	if aspect:
		ax.set_aspect(aspect)  # Plot the figure
	# Note on Tick Labels:
	#   If a group of subplots are put together which share a x-axis, the rotation may not work. By
	#   getting the tick labels generated by matplotlib as a set of label objects, they can be
	#   looped through and have their settings individually fixed. This appears to be the only way
	#   to have the shared axis labels formated properly. The labels are also adjusted closer to
	#   the axis for esthetics. --Warren E. Black

	# The plots tick labels will not be properly accessible until the figure is "drawn", once the
	# command below is run, ax.get_ticklabel() will actually work properly.
	plt.draw()
	# Check to see if the ytick labels need to be rotated if the rotate argument was not passed.
	# This doesn't work if the figure is being placed into a subplot (i.e., is not a standalone
	# figure). Don't know why...
	ax = _format_tick_labels(ax, rotateticks)

	# Plot colorbar, the tick locations and labels were generated above using get_vlimdata
	if cbar:
		# Set-up axis to place colorbar into
		# cax = grid.cbar_axes[0]
		# Plot the colorbar
		if drill_plot_format:
			cbar = plt.colorbar(lc, cax=cax, ticks=ticklocs)
		else:
			cbar = fig.colorbar(plot, cax=cax, ticks=ticklocs)
		# Configure the color bar
		cbar.ax.set_yticklabels(ticklabels, ha='left')
		cbar.ax.tick_params(axis='y', pad=2)
		if cbar_label is not None:
			cbar.set_label(cbar_label, ha='center', va='top', labelpad=2)
	if axis_xy is None:
		axis_xy = Parameters['plotting.axis_xy_spatial']
	format_plot(ax, grid=grid, axis_xy=axis_xy, xlim=xlim, ylim=ylim)
	# Export figure
	if output_file or ('pdfpages' in out_kws):
		export_image(output_file, **out_kws)
	# Return whats needed
	if return_cbar:
		return ax, cbar
	if return_plot:
		return plot
	return ax


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

import six
import os
from collections import MutableMapping
import warnings

# Validation functions for parameters
#----------------------------------------------------------------------------------------------------------
def _validate_string(s, accept_none = False):
	"""
	 A validation method to convert input s to string or raise error if it is not convertable
	"""
	if s is None and accept_none :
		return None
	try:
		if isinstance(s,list):
			return [six.text_type(item) for item in s]
		elif isinstance(s,dict):
			return dict((six.text_type(key), six.text_type(value)) for key, value in s.items())
		else:
			return six.text_type(s)
	except ValueError:
		raise ValueError('Could not convert "%s" to string' % s)

def _validate_string_or_None(s):
	return _validate_string(s, accept_none=True)

def _validate_float(s, accept_none = False):
	"""
	A validation method to convert input s to float or raise error if it is not convertable
	"""

	if s is None and accept_none :
		return None
	try:
		return float(s)
	except ValueError:
		raise ValueError('Could not convert "%s" to float' % s)

def _validate_float_or_None(s):
	return _validate_float(s, accept_none=True)

def _validate_int(s, accept_none = False):
	"""
	A validation method to convert input s to int or raise error if it is not convertable
	"""
	if s is None and accept_none :
		return None
	try:
		return int(s)
	except ValueError:
		raise ValueError('Could not convert "%s" to int' % s)

def _validate_int_or_None(s):
	return _validate_int(s, accept_none=True)

def _validate_color(s):

	"""
	A validation method to check s is a valid matplotlib color
	"""

	from matplotlib.colors import is_color_like

	try:
		if s.lower() == 'none':
			return 'none'
	except AttributeError:
		pass

	if isinstance(s, six.string_types):
		if len(s) == 6 or len(s) == 8:
			stmp = '#' + s
			if is_color_like(stmp):
				return stmp

	if is_color_like(s):
		return s

def _validate_list(s, accept_none=False):
	"""
	A validation method to convert input s to a list or raise error if it is not convertable
	"""
	if s is None and accept_none:
		return None
	try:
		return list(s)
	except ValueError:
		raise ValueError('Could not convert input to list')

def _validate_list_or_None(s):
	return _validate_list(s, accept_none=True)

def _validate_list_int(s, accept_none=False):
	"""
	A validation method to convert input s to int or list and raise error if it is not convertable
	"""
	if s is None and accept_none:
		return None
	try:
		return int(s)
	except (ValueError, TypeError):
		try:
			return list(s)
		except ValueError:
			raise ValueError('Could not convert input to int or list')

def _validate_list_int_or_None(s):
	return _validate_list_int(s, accept_none=True)

def _validate_bool(s):
	"""
	A validation method to convert input s to boolean or raise error if it is not convertable
	"""
	try:
		return bool(s)
	except ValueError:
		raise ValueError('Could not convert input to boolean')

def _validate_dict_or_string(s):
	"""
	A validation method to convert input s to string or dictionary and raise error if it is not convertable
	"""
	if isinstance(s, str) or isinstance(s, dict):
		return s
	try:
		return str(s)
	except ValueError:
		raise ValueError('Could not convert input to dict or string')

def _validate_dict(s, accept_none=False):
	"""
	A validation method to check if the input s is a dictionary otherwise raise error if it is not convertable
	"""
	if s is None and accept_none:
		return None
	if isinstance(s, dict):
		return s
	else:
		raise ValueError('{} is not a dictionary!'.format(s))

def _validate_dict_or_None(s):
	return _validate_dict(s, accept_none=True)

def _validate_kde_or_color(s):
	'''
	A validation method to return a valid color arg if not KDE
	'''

	from matplotlib.colors import is_color_like

	try:
		if s.lower() == 'kde':
			return 'kde'
	except AttributeError:
		pass
	try:
		if s.lower() == 'none':
			return 'none'
	except AttributeError:
		pass

	if isinstance(s, six.string_types):
		if len(s) == 6 or len(s) == 8:
			stmp = '#' + s
			if is_color_like(stmp):
				return stmp

	if is_color_like(s):
		return s

def _validate_griddef_or_None(s):
	"""
	A validation method to convert input s to pygeostat grid def and raise an error if not convertable.
	"""
	import pygeostat as gs
	if s is None:
		return None
	elif isinstance(s, str):
		try:
			return gs.GridDef(s)
		except AssertionError:
			raise ValueError('{} should be a pygeostat.GridDef object or griddef string!'.format(s))
	elif isinstance(s, gs.GridDef):
		return s
	else:
		raise ValueError('{} should be a pygeostat.GridDef object or griddef string!'.format(s))

#----------------------------------------------------------------------------------------------------------

# Pygeostat configurations file
_params_cfg_file = os.path.join(os.path.expanduser('~'),r".Pygeostat\Parameters.json")

# Writing/loading parameters to a json file
#----------------------------------------------------------------------------------------------------------
def _write_pardict(filename, dictionary, section, description=None):
	import json
	from .data.grid_definition import GridDef

	if os.path.isfile(filename):
		with open(filename, 'r') as cfg_file:
			dict_main = dict(json.load(cfg_file))
	else:
		dict_main = {}
	if section in dict_main.keys():
		dict_main.pop(section, None)
	
	section_dict={}
	for key in list(sorted(dictionary.keys())):
		if 'griddef' in str(key).lower() and dictionary[key] is not None:
			value = str(dictionary[key])
		else:
			value = dictionary[key]
		section_dict.update({key: {'Value': value, 'Description':None if description is None else description[key]}})
	
	dict_main.update({section:section_dict})

	with open(filename, 'w') as cfg_file:
		json.dump(dict_main, cfg_file)

def _parse_pardict(filename, section):
	import json
	if os.path.isfile(filename):
		try:
			with open(filename, 'r') as fp:
				dict_main = dict(json.load(fp))
		except:
			print('Unable to open the provided parameter file: {}'.format(filename))
			return {}
	else:
		print('Unable to load the system defaults as it does no exist! Use gs.Programs.set_systems_defaults to create one.')
		return {}
	# Allowing no values
	if section not in dict_main.keys():
		return {}
		
	new_dict = {}
	input_dict = dict_main[section]
	for key in input_dict:
		new_dict.update({key: input_dict[key]['Value']})

	return new_dict

# Parameter list
#---------------------------------------------------------------------------------------------------
default_parameters = {
	'config.verbose': [True, _validate_bool],
	'config.autoload.parameters': [True, _validate_bool],
	'config.autoload.plot_style': [False, _validate_bool],
	'config.use_shipped_executable': [True, _validate_bool],
	'config.nprocess': [4, _validate_int_or_None],
	'config.getpar': [False, _validate_bool],
	'config.ignore_mpl_warnings': [True, _validate_bool],
	'plotting.unit': ['m', _validate_string_or_None],
	'plotting.xname': ['Easting', _validate_string_or_None],
	'plotting.yname': ['Northing', _validate_string_or_None],
	'plotting.zname': ['Elevation', _validate_string_or_None],
	'plotting.xabbrev': ['E', _validate_string_or_None],
	'plotting.yabbrev': ['N', _validate_string_or_None],
	'plotting.zabbrev': ['Elev', _validate_string_or_None],
	'plotting.gammasize': [3.0, _validate_float],
	'plotting.lagname': ['Lag Distance', _validate_string_or_None],
	'plotting.rotateticks': [[0, 0], _validate_list_or_None],
	'plotting.nticks': [None, _validate_list_int_or_None],
	'plotting.grid': [False, _validate_bool],
	'plotting.axis_xy': [False, _validate_bool],
	'plotting.axis_xy_spatial': [False, _validate_bool],
	'plotting.stat_fontsize': [None, _validate_float_or_None],
	'plotting.stat_ha': ['right', _validate_string],
	'plotting.stat_linespacing': [1.0, _validate_float],
	'plotting.roundstats': [True, _validate_bool],
	'plotting.sigfigs': [2, _validate_int],
	'plotting.cmap': ['viridis', _validate_string],
	'plotting.cmap_cat': ['tab20', _validate_dict_or_string],
	'plotting.assumecat': [11, _validate_int],
	'plotting.log_lowerval': [0.0001, _validate_float],
	'plotting.histogram_plot.facecolor': ['.9', _validate_color],
	'plotting.histogram_plot.edgecolor': ['k', _validate_color],
	'plotting.histogram_plot.edgeweight': [None, _validate_float_or_None],
	'plotting.histogram_plot.cdfcolor': ['.5', _validate_color],
	'plotting.histogram_plot.histbins': [15, _validate_int_or_None],
	'plotting.histogram_plot.stat_blk': ['all', _validate_string],
	'plotting.histogram_plot.stat_xy': [(0.95, 0.95), _validate_list],
	'plotting.histogram_plot.stat_xy_cdf': [(0.95, 0.05), _validate_list],
	'plotting.histogram_plot_simulation.refclr': ['C3', _validate_color],
	'plotting.histogram_plot_simulation.simclr': ['0.2', _validate_color],
	'plotting.histogram_plot_simulation.alpha': ['0.5', _validate_float_or_None],
	'plotting.location_plot.s': [20, _validate_float_or_None],
	'plotting.location_plot.c': ['.4', _validate_color],
	'plotting.location_plot.lw': ['5', _validate_float_or_None],
	'plotting.scatter_plot.s': [None, _validate_float_or_None],
	'plotting.scatter_plot.c': ['kde', _validate_kde_or_color],
	'plotting.scatter_plot.cmap': ['viridis', _validate_string],
	'plotting.scatter_plot.alpha': [1.0, _validate_float_or_None],
	'plotting.scatter_plot.stat_blk': ['pearson', _validate_string],
	'plotting.scatter_plot.stat_xy': [(0.95, 0.05), _validate_list],
	'plotting.variogram_plot.color': ['.5', _validate_color],
	'plotting.variogram_plot.ms': [6.0, _validate_float],
	'plotting.variogram_plot_simulation.refclr': ['C3', _validate_color],
	'plotting.variogram_plot_simulation.simclr': ['0.2', _validate_color],
	'plotting.variogram_plot_simulation.alpha': ['0.5', _validate_float_or_None],
	'plotting.vplot.colors': [['C0', 'C1', 'C2'], _validate_string],
	'data.tmin': [-998.0, _validate_float_or_None],
	'data.null': [-999.0, _validate_float_or_None],
	'data.legacy_null': [(-999, -998, -99, -98), _validate_list_or_None],
	'data.fix_legacy_null': [False, _validate_bool],
	'data.null_vtk': [False, _validate_float_or_None],
	'data.griddef': [None, _validate_griddef_or_None],
	'data.nreal': [None, _validate_int_or_None],
	'data.dh': [None, _validate_string_or_None],
	'data.ifrom': [None, _validate_string_or_None],
	'data.ito': [None, _validate_string_or_None],
	'data.x': [None, _validate_string_or_None],
	'data.y': [None, _validate_string_or_None],
	'data.z': [None, _validate_string_or_None],
	'data.cat': [None, _validate_string_or_None],
	'data.catdict': [None, _validate_dict_or_None],
	'data.weights': [None, _validate_string_or_None],
	'data.io.pandas_engine': ['python', _validate_string],
	'data.write.python_floatfmt': ['%.6f', _validate_string],
	'data.write_vtk.vdtype': ['float64', _validate_string],
	'data.write_vtk.cdtype': ['float64', _validate_string],
}

# Description of the parameters
default_parameter_descriptions = {
	'config.verbose': ('If `true`, mentions when the `gsParams` are updated automatically '
					   'when pygeostat is loaded. Useful for interactive sessions'),
	'config.autoload.parameters': ('If `True` the `gsParams` configurations found in\n'
								 '``%USER%/.gsParams`` are parsed and loaded when pygeostat loads'),
	'config.autoload.plot_style': ('If `True` the `gsPlotStyle` configurations found in\n'
									'``%USER%/.gsParams`` are parsed and loaded when pygeostat loads'),
	'config.use_shipped_executable': ('If `True` adds the shipped executable files to the path environmental variable'),
	'config.nprocess': ('The number of parallel processes to run functions that provide that '
						'functionality'),
	'config.getpar': ('If True, getpar=True when calling pygeostat.Program() with getpar=None'),
	'config.ignore_mpl_warnings': 'If True, matplotlib warnings will be ignored!',
	'plotting.unit': ('The unit (str) that appears in spatial/map/variogram plots '
					  'e.g., m or ft. Use None if no unit should be displayed.'),
	'plotting.xname': ('Name (str) that is used for the x coordinate label for '
					   'map/cross-section plots, e.g., Easting or X'),
	'plotting.yname': ('Name (str) that is used for the y coordinate label for '
					   'map/cross-section plots, e.g., Northing or Y'),
	'plotting.zname': ('Name (str) that is used for the z coordinate label for '
					   'map/cross-section plots, e.g., Elevation or Z'),
	'plotting.xabbrev': ('Abbreviated name (str) that is used for the x coordinate '
						 'label for map/cross-section plots, e.g., Easting or X'),
	'plotting.yabbrev': ('Abbreviated name (str) that is used for the y coordinate '
						 'label for map/cross-section plots, e.g., Northing or Y'),
	'plotting.zabbrev': ('Abbreviated name (str) that is used for the z coordinate '
						 'label for map/cross-section plots, e.g., Elevation or Z'),
	'plotting.gammasize': ('The size of gamma symbols in variogram plots are '
						   'Matplotlib.rcParams["font.size"] multiplied with this value.'),
	'plotting.lagname': ('Name (str) to use on the x-axis of variogram plots, '
						 'e.g., Lag or h'),
	'plotting.rotateticks': ('If None, the pygeostat tickoverlap attempts to optimize tick, '
							 'label angles to minimize overlap in the absence of a kwarg., '
							 'If [xangle, yangle] (e.g., [0, 0], then these angles are used '
							 'in the absence of a kwarg.'),
	'plotting.nticks': ('Specify the target number of ticks for plotting, as an integer, or '
						'tuple/list of integers specifying the number of ticks on the x and y '
						'axis, respectively'),
	'plotting.grid': ('If True, a grid is plotted in pygeostat ploting function unless, '
					  'overridden by the associated kwarg'),
	'plotting.axis_xy': ('Converts plotting axes to GSLIB-style visibility (only left and '
						 'bottom visible) if axis_xy is True. See plotting.axis_xy_spatial '
						 'for the default setting of functions such as slice_plot and location_plot.'),
	'plotting.axis_xy_spatial': ('axis_xy setting that is specific to spatial plotting functions '
								 'such as location_plot and slice_plot. Provided since many users '
								 'prefer axis_xy to be applied to all plots (via '
								 'plotting.axis_xy) other than spatial plots. '),
	'plotting.stat_fontsize': ('Font size of statistics in pygeosat plots. If None, then '
							   'the font size matches rcParams[font.size]. If a fraction, '
							   'then the font size is rcParams[font.size]*stat_fontsize. '
							   'If greater than or equal to 1, then stat_fontsize is the '
							   'font size.'),
	'plotting.stat_linespacing': ('The line spacing to use for statistics blocks in plots. '
								  'Default is 1.0'),
	'plotting.stat_ha': ('The horizontal alignment of statistics in plots, which should be one of '
						 '"right", "left" or "center"'),
	'plotting.roundstats': ('Display a set number of significant figures (False) or digits (True) '
							'for statistics in plots.'),
	'plotting.sigfigs': ('Significant figures (roundstats=False) or digits (roundstats=True) '
						 'for statistics in plots.'),
	'plotting.cmap': ('Matplotlib colormap object or a registered Matplotlib colormap, which is '
					  'used for continuous variables'),
	'plotting.cmap_cat': ('Matplotlib colormap object or a registered Matplotlib colormap, which '
						  'is used for categorical variables'),
	'plotting.assumecat': ('When executing plotting functions, data is assumed to be categorical '
						   'for the purposes of selecting colormaps, if less than this number '
						   'of unique values are found'),
	'plotting.log_lowerval': ('For a log axis, specify the default lower value that replaces a '
							  '0.0 or negative valued data. Applies to both axes.'),
	'plotting.histogram_plot.facecolor': 'Color of the histogram faces for the histogram_plot function.',
	'plotting.histogram_plot.edgecolor': 'Color of the histogram edges for the histogram_plot function.',
	'plotting.histogram_plot.edgeweight': '`lw` of the histogram edges for the histogram_plot function.',
	'plotting.histogram_plot.cdfcolor': 'Color of the CDF for the histogram_plot function.',
	'plotting.histogram_plot.histbins': 'Number of bins for a histogram (not CDF) calculation.',
	'plotting.histogram_plot.stat_blk': ('Default stat_blk setting, which is either a string code, such '
								  'as all or minimal, or a list of valid statistic names'),
	'plotting.histogram_plot.stat_xy': ('Location of the histogram stats location as a tuple '
								 'of (xloc, yloc), where the locations should be between 0 and 1.'),
	'plotting.histogram_plot.stat_xy_cdf': ('Location of the CDF stats location as a tuple of  '
									 '(xloc, yloc), where the locations should be 0 to 1.'),
	'plotting.histogram_plot_simulation.refclr': 'Color of the reference CDF',
	'plotting.histogram_plot_simulation.simclr': 'Color of the realization CDFs',
	'plotting.histogram_plot_simulation.alpha': 'Transparency of the realizations CDFs',
	'plotting.location_plot.s': ('Size of scatter in location_plot (and slice_plot), which is based on '
						  'Matplotlib.rcParams if None'),
	'plotting.location_plot.c': ('Color of scatter in location_plot (and slice_plot), which is a valid Matplotlib '
						  'color specifier'),
	'plotting.location_plot.lw': ('Line width for location plot if the orient is xz or yz and the drill hole id is provided/inferred'),
	'plotting.scatter_plot.s': ('Size of scatter in scatter_plot, which is based on Matplotlib.rcParams if '
						   'None'),
	'plotting.scatter_plot.c': ('Color of scatter in scatter_plot, which is either a valid Matplotlib '
						   'color specifier, or the "KDE" string. KDE leads to calculation of '
						   'the kernel density estimate at each scatter location'),
	'plotting.scatter_plot.cmap': ('Matplotlib colormap object or a registered Matplotlib colormap '
							  'which overrides plotting.cmap'),
	'plotting.scatter_plot.alpha': ('Alpha transparency of scatter in scatter_plot, which should be '
							   'between 0 and 1.'),
	'plotting.scatter_plot.stat_blk': ('Statistics that are plotted in scatter_plot, which should be '
								  'either "all" or a list that may contains the strings: '
								  '["count", "pearson", "spearmanr"]'),
	'plotting.scatter_plot.stat_xy': 'A 2-list that provides the x/y location of the statistics block',
	'plotting.variogram_plot.color': 'Color for variogram_plot ',
	'plotting.variogram_plot.ms': 'Marker size for variogram_plot (e.g. experimental dot size) ',
	'plotting.variogram_plot_simulation.refclr': 'Color of the reference variogram',
	'plotting.variogram_plot_simulation.simclr': 'Color of the realization variograms',
	'plotting.variogram_plot_simulation.alpha': 'Transparency of the realizations variograms',
	'plotting.vplot.colors': ('Colors of the major, minor and vertical direction variograms, '
							  'which is used by the Variogram.plot function. May be a 3-list, '
							  'or a color pallete in gs.avail_palettes'),
	'data.tmin': ('Values less than this number (float or int) are assigned NaN on '
				  'import by all io read functions, e.g., -98 or -998. Using -1.0e21 '
				  'or False disables this functionality.'),
	'data.null': ('NaN values are assigned this number (float or int) by io write '
				  'functions (excluding write_vtk), e.g., -99, -999, or -999.0. Using '
				  'None disables the functionality, but will likely lead to issues '
				  'in the case of subsequent GSLIB operations.'),
	'data.null_vtk': ('NaN values are assigned this number (float or int) by the '
					  'write_vtk function, e.g., -99, -999, or -999.0. Using None '
					  'disables the functionality, which is recommended since NaN '
					  'values are handled by Paraview.'),
	'data.legacy_null': ('Nulls used by default in some GSLIB programs that can be fixed if '
						 'encountered'),
	'data.fix_legacy_null': ('Automatically fix legacy null values if encountered and '
							 'trimming null values'),
	'data.griddef': ('When initializing a DataFile, this will be used as '
					 'DataFile.GridDef if GridDef.count() matches DataFile.shape[0]. '
					 'A pygeostat.GridDef object or valid gridstr/gridarr may be used '
					 'for intitialization.'),
	'data.nreal': ('The number of realizations for modeling (once implemented) '
				   'and model checking.'),
	'data.dh': ('When initializing a DataFile, this str will be used as '
				'DataFile.dh if it is in DataFile.columns, e.g., WellID'),
	'data.ifrom': ('When initializing a DataFile, this str will be used as '
				   'DataFile.ifrom if it is in DataFile.columns, e.g., From'),
	'data.ito': ('When initializing a DataFile, this str will be used as '
				 'DataFile.ito if it is in DataFile.columns, e.g., To'),
	'data.x': ('When initializing a DataFile, this st will be used as '
			   'DataFile.x if it is in DataFile.columns, e.g., Easting'),
	'data.y': ('When initializing a DataFile, this str will be used as '
			   'DataFile.y if it is in DataFile.columns, e.g., Northing'),
	'data.z': ('When initializing a DataFile, this str will be used as '
			   'DataFile.z if it is in DataFile.columns, e.g., Elevation'),
	'data.cat': ('When initializing a DataFile, this str will be used as '
				 'DataFile.cat if it is in DataFile.columns, e.g., Facies'),
	'data.catdict': ('When initializing a DataFile, this dictionary will be used as '
					 'DataFile.catdict (if catdict.keys() matches DataFile.cat codes). This '
					 'dictionary is formaatted as {catcodes: catnames}.'),
	'data.weights': ('When initializing a DataFile, this str or list will be used as '
				 'DataFile.wts if wts is in DataFile.columns, e.g., DeclusteringWeight'),
	'data.io.pandas_engine': 'Default engine to use for pandas read, either `python` or `c`',
	'data.write.python_floatfmt': ('A format string to use for `gslib` fltypes (written with '
								   'pandas)'),
	'data.write_vtk.vdtype': ('Precision of variables when writing to VTK. Must be a valid '
							  'numpy specifier, such as "float32", "float64", "int8", etc.'),
	'data.write_vtk.cdtype': ('Precision of coordinates when writing to VTK. Must be a valid '
							  'numpy specifier, such as "float32", "float64", etc.'),
}

#----------------------------------------------------------------------------------------------------------

# Plotstyles
#----------------------------------------------------------------------------------------------------------
import matplotlib as mpl

nochanges = {}

darkcontent = {'axes.axisbelow': True, 'axes.edgecolor': 'white', 'axes.facecolor': 'darkgray',
			   'axes.grid': False, 'axes.labelcolor': 'white', 'axes.labelsize': 8,
			   'axes.linewidth': 0.5, 'axes.titlesize': 12,
			   'figure.figsize': (5, 5), 'figure.subplot.bottom': 0.125,
			   'figure.facecolor': 'black', 'figure.edgecolor': 'black',
			   'font.family': 'Calibri', 'font.size': 8, 'font.weight': 400, 'text.color': 'white',
			   'grid.color': 'lightgray', 'grid.linestyle': '-', 'grid.linewidth': 0.5,
			   'legend.fontsize': 8.0, 'legend.frameon': False, 'legend.numpoints': 1,
			   'legend.scatterpoints': 1,
			   'lines.linewidth': 1.5, 'lines.markeredgewidth': 0, 'lines.markersize': 7,
			   'lines.solid_capstyle': 'round', 'patch.linewidth': 0.3,
			   'pdf.fonttype': 3, 'ps.fonttype': 3, 'ps.useafm': True,
			   'xtick.color': 'white', 'xtick.direction': 'out', 'xtick.labelsize': 8,
			   'xtick.major.pad': 3, 'xtick.major.width': 1, 'xtick.major.size': 0,
			   'xtick.minor.width': 0.5, 'xtick.minor.size': 0, 'ytick.color': 'white',
			   'ytick.direction': 'out', 'ytick.labelsize': 8, 'ytick.major.pad': 3,
			   'ytick.major.width': 1, 'ytick.major.size': 0, 'ytick.minor.width': 0.5,
			   'ytick.minor.size': 0}

ccgpaper = {'axes.axisbelow': True, 'axes.edgecolor': 'black', 'axes.facecolor': 'white',
			'axes.grid': False, 'axes.labelcolor': 'black', 'axes.labelsize': 8,
			'axes.linewidth': 0.5, 'axes.titlesize': 10,
			'figure.figsize': (8, 8), 'figure.subplot.bottom': 0.125, 'figure.facecolor': 'white',
			'figure.edgecolor': 'black',
			'font.family': 'Calibri', 'font.size': 12, 'font.weight': 400, 'text.color': 'black',
			'grid.color': 'lightgray', 'grid.linestyle': '-', 'grid.linewidth': 0.5,
			'legend.fontsize': 8.0, 'legend.frameon': False, 'legend.numpoints': 1,
			'legend.scatterpoints': 1,
			'lines.linewidth': 1.5, 'lines.markeredgewidth': 0, 'lines.markersize': 7,
			'lines.solid_capstyle': 'round', 'patch.linewidth': 0.3,
			'pdf.fonttype': 3, 'ps.fonttype': 3, 'ps.useafm': True,
			'xtick.color': 'black', 'xtick.direction': 'out', 'xtick.labelsize': 11,
			'xtick.major.pad': 3, 'xtick.major.width': 1, 'xtick.major.size': 0,
			'xtick.minor.width': 0.5, 'xtick.minor.size': 0, 'ytick.color': 'black',
			'ytick.direction': 'out', 'ytick.labelsize': 11, 'ytick.major.pad': 3,
			'ytick.major.width': 1, 'ytick.major.size': 0, 'ytick.minor.width': 0.5,
			'ytick.minor.size': 0}

mpldefault = {} # Avoid loading all rcParams to minimize deprectation warnings
mpldefault.update((key,mpl.rcParams[key]) for key in ccgpaper.keys())

presentation = {'axes.axisbelow': True, 'axes.edgecolor': 'black', 'axes.facecolor': 'white',
				'axes.grid': False, 'axes.labelcolor': 'black', 'axes.labelsize': 8,
				'axes.linewidth': 0.5, 'axes.titlesize': 8, 'figure.figsize': (8, 8),
				'figure.subplot.bottom': 0.125, 'figure.facecolor': 'white',
				'figure.edgecolor': 'black',
				'font.family': 'Calibri', 'font.size': 12, 'font.weight': 400,
				'text.color': 'black', 'grid.color': 'lightgray', 'grid.linestyle': '-',
				'grid.linewidth': 0.5, 'legend.fontsize': 8.0, 'legend.frameon': False,
				'legend.numpoints': 1, 'legend.scatterpoints': 1, 'lines.linewidth': 1.5,
				'lines.markeredgewidth': 0, 'lines.markersize': 5, 'lines.solid_capstyle': 'round',
				'patch.linewidth': 0.3, 'pdf.fonttype': 3, 'ps.fonttype': 3, 'ps.useafm': True,
				'xtick.color': 'black', 'xtick.direction': 'out', 'xtick.labelsize': 8,
				'xtick.major.pad': 3, 'xtick.major.width': 1, 'xtick.major.size': 0,
				'xtick.minor.width': 0.5, 'xtick.minor.size': 0, 'ytick.color': 'black',
				'ytick.direction': 'out', 'ytick.labelsize': 8, 'ytick.major.pad': 3,
				'ytick.major.width': 1, 'ytick.major.size': 0, 'ytick.minor.width': 0.5,
				'ytick.minor.size': 0}

jdplots = {'axes.axisbelow': True, 'axes.edgecolor': 'k', 'axes.facecolor': 'white',
		   'axes.grid': False, 'axes.labelcolor': 'k', 'axes.labelsize': 'medium',
		   'axes.linewidth': 1.0, 'axes.titlesize': 'large',
		   'figure.figsize': (7, 5), 'figure.subplot.bottom': 0.125, 'figure.facecolor': 'white',
		   'figure.edgecolor': 'white',
		   'font.family': 'sans-serif', 'font.size': 12.0, 'font.weight': 'medium',
		   'grid.color': 'lightgray', 'grid.linestyle': '-', 'grid.linewidth': 0.5,
		   'legend.fontsize': 'medium', 'legend.frameon': False, 'legend.numpoints': 1,
		   'legend.scatterpoints': 1,
		   'lines.linewidth': 1.0, 'lines.markeredgewidth': 0.25, 'lines.markersize': 6,
		   'lines.solid_capstyle': 'round', 'patch.linewidth': 0.3,
		   'pdf.fonttype': 3, 'ps.fonttype': 3, 'ps.useafm': False,
		   'text.color': 'k', 'xtick.color': 'k', 'xtick.direction': 'in',
		   'xtick.labelsize': 'medium', 'xtick.major.pad': 4, 'xtick.major.width': 0.5,
		   'xtick.major.size': 4, 'xtick.minor.width': 0.5, 'xtick.minor.size': 0,
		   'ytick.color': 'k', 'ytick.direction': 'in', 'ytick.labelsize': 'medium',
		   'ytick.major.pad': 4, 'ytick.major.width': 0.5, 'ytick.major.size': 4,
		   'ytick.minor.width': 0.5, 'ytick.minor.size': 0}

# Create alternate styles of the default for different font sizes
fontparms = ['axes.labelsize', 'axes.titlesize', 'legend.fontsize', 'font.size', 'xtick.labelsize',
			 'ytick.labelsize']

# Define a dictionary pointing to the right data
plot_styles = {'nochanges': nochanges,
				'darkcontent': darkcontent,
				'ccgpaper': ccgpaper,
				'presentation': presentation,
				'mpldefault': mpldefault,
				'jdplots': jdplots}


class Parameters(MutableMapping, dict):

	"""
	A dictionary object for defining pygeostat default parameters, which is intended
	to add convenience to modeling workflows where common parameters must otherwise be
	repeated in function calls. An instance of this object is initialized as Parameters when
	pygeostat is imported. Default parameters are implemented as thoughtfully
	as possible, but may be modified by users once to suit their requirements. Keyword
	arguments for functions override any Parameters defaults for the function call itself,
	without modifying the defaults themselves.

	Examples of modifying plotting parameters include:

		>>> gs.Parameters['plotting.xname'] = 'Easting'
		>>> gs.Parameters['plotting.lagname'] = 'h'
		>>> gs.Parameters['plotting.unit'] = 'm'

	Examples of data parameters include:

		>>> gs.Parameters['data.dh'] = 'WellID'
		>>> gs.Parameters['data.griddef'] = pygeostat.GridDef()
		>>> gs.Parameters['data.x'] = 'Easting'

	The object performs validation on any modificaitons to the dictionary. If users are confused
	about the application of any defaults, a describe function allows for individual parameters
	(e.g., Parameters.describe('plotting.xname') or the entire dictionary (e.g., Parameters.describe()
	to be printed to the console.

	Attributation to Matplotlib:

		The template for the Parameters class is inspired by adapted from the
		Matplotlib rcParams class and related functionality, as it provides a template for the
		approach to setting defaults to a Python instance.
	"""
	# get validation items
	validate = dict((key, converter) for key, (_, converter) in six.iteritems(default_parameters))
	

	def __init__(self, *args, **kwargs):
		self.update(*args, **kwargs)
		self._section_name = 'PygeostatParameters'
		self._default_dict = dict((key, parameter) for key, (parameter, _) in six.iteritems(default_parameters))
		self._default_descriptions = default_parameter_descriptions


	def __setitem__(self, key, val):
		try:
			cval = self.validate[key](val)
			dict.__setitem__(self, key, cval)
		except ValueError as ve:
				raise ValueError("Key %s: %s" % (key, str(ve)))
		except:
			warnings.warn('%s is not a valid parameter! See Parameters.keys() for a '
						   'list of valid parameters.' % (key,))
	
	def __getitem__(self, key):
		val = dict.__getitem__(self, key)
		return val

	def __iter__(self):
		"""
		Yield sorted list of keys.
		"""
		for k in sorted(dict.__iter__(self)):
			yield k

	def __repr__(self):
		import pprint
		class_name = self.__class__.__name__
		indent = len(class_name) + 1
		repr_split = pprint.pformat(dict(self), indent=1,
									width=80 - indent).split('\n')
		repr_indented = ('\n' + ' ' * indent).join(repr_split)
		return '{0}({1})'.format(class_name, repr_indented)

	def __str__(self):
		return '\n'.join('{}: {}'.format(k, v) for k, v in sorted(self.items()))

	def __delitem__(self, key):
		raise KeyError('Not allowed to remove a paramater!')

	def find_all(self, pattern):
		"""
		Return the subset of this Parameters dictionary whose keys match,
		using :func:`re.search`, the given ``pattern``.

		.. note::

			Changes to the returned dictionary are *not* propagated to
			the parent Parameter dictionary.
		"""
		import re
		pattern_re = re.compile(pattern)
		return dict([(key, value) for key, value in self.items() if pattern_re.search(key)])

	def restore_defaults(self):
		"""
		Restore to the original dictionary of defaults values for the current instance
		"""
		self.update([(key, default_value) for key, default_value in self._default_dict.items()])


	def describe(self, key=None, verbose=True):
		"""
		Print a description of an individual parameter in the
		pygeostat parameter collection, as specified by its dictionary key. If key
		is None, all parameter descriptions are printed.

		Parameters:
			key(str) : dictionary key within Parameters

		"""
		if key is None:
			# Print the entire dictionary description
			for iterkey in iter(self):
				print(iterkey + ':\n' + self._default_descriptions[iterkey] + '\n')
		else:
			if key not in iter(self):
				raise KeyError('%s is not a valid parameter. See pygeostat.Parameters.keys() for a '
							   'list of valid parameters.' % (key,))
			if verbose:
				print(key + ':\n' + self._default_descriptions[key])

	def save(self, filename):
		"""
		Save the current pygeostat configurations to the given file
		"""
		_write_pardict(filename, self, self._section_name, description=self._default_descriptions)

	def load(self, filename):
		"""
		Load the pygeostat configurations from a given file
		"""
		import os
		assert os.path.isfile(filename), "ERROR: `{}` does not exist".format(filename)
		newdict = _parse_pardict(filename, self._section_name)
		self.update(newdict)

	def reset_systemdefault(self):
		"""
		Overwrite the system defaults stored in ``%USER%/.Pygeostat`` with the default parameters of
		pygeostat
		"""
		_write_pardict(_params_cfg_file, self._default_dict, self._section_name, description=self._default_descriptions)


	def set_systemdefault(self):
		"""
		Write the current configuration from the current instance i.e. ``self`` to the system defaults in ``%USER%/.Pygeostat``
		"""
		print("Writing current `{}` to system default at: {}".format(self._section_name, _params_cfg_file))
		main_dir = os.path.dirname(_params_cfg_file)
		if not os.path.exists(main_dir):
			os.makedirs(main_dir)
		self.save(_params_cfg_file)


	def get_systemdefault(self, confirm_autoload=False):
		"""
		Get the Parameters stored in the system defaults in ``%USER%/.Pygeostat``. ``confirm_autoload``
		is passed only on the very first call on module init.
		"""
		if os.path.isfile(_params_cfg_file):
			newdict = _parse_pardict(_params_cfg_file, section=self._section_name)
			
			if newdict: # If dictionary is not empty
				if confirm_autoload:
					if newdict.get("config.autoload.parameters", False):
						try:
							self.update(newdict)
							if newdict.get("config.verbose", False):
								print("Loading default Pygeostat Parameters from {}".format(_params_cfg_file))
						except:
							print('Unable to load the system defaults due to version issues.\nThe current system defaults setting was removed. Please try Pygeostat.Parameters.set_systemdefault()')
							os.remove(_params_cfg_file)
				else:
					self.update(newdict)
					if newdict.get("config.verbose", False):
						print("Loading default Pygeostat Parameters from {}".format(_params_cfg_file))


class PlotStyle(Parameters):
	"""
	Manages a set of pygeostat default plotting styles VIA ``mpl.rcParams``.

	Parameters:
		plot_style (str): One of: "darkcontent", "ccgpaper", "presentation", "mpldefault", "jdplots"
		custom_dict (dict): Dictionary of custom mpl.rcParams

	Note:
		Default behavior is to use the ``mpl`` defaults. ``PlotStyle.set_style()`` changes the defaults
		stored in this class for the current session

	Examples:

		The main instance of this class used by all plotting functions is accessible with
		``gs.PlotStyle``. The only things that the user should require from this instance is to
		either save the configured plot style to a given file:

		>>> gs.PlotStyle.save("plottingstyles.cfg")

		Load a saved configuration from a saved output:

		>>> gs.PlotStyle.load("plottingstyles.cfg")

		or to write system defaults, or load system defaults:

		>>> gs.PlotStyle.set_systemdefault()
		>>> gs.PlotStyle.get_systemdefault()

		one can optionally restore the defaults for the current session with:

		>>> gs.PlotStyle.restore_defaults()

		To set ``ccgpaper`` as the system default, the following is required:

		>>> gs.PlotStyle.set_style("ccgpaper")  # this is now the session default
		>>> gs.gsPlotStyle.set_systemdefault()

		``gsPlotStyle`` params are autoloaded if ``Parameters["config.autoload.gsplotstyle"] == True``

		>>> gs.Parameters["config.autoload.gsplotstyle"] = True
		>>> gs.Parameters.set_systemdefault()

	"""

	import matplotlib as mpl

	def __init__(self, plot_style, custom_dict=None):
		if plot_style in (None, False):
			plot_style = "nochanges"
		self.style_string = plot_style

		self._default_style = 'mpldefault'
		self._default_dict = plot_styles[self._default_style]

		self._default_descriptions = None
		self._section_name = 'PlotStyle'

		self.update(self._generate(plot_style, custom_dict))
	
	def set_style(self, plot_style, custom_dict=None):
		'''
		Sets the plot style based on the available styles
		'''
		dict.clear(self)
		if plot_style in (None, False):
			plot_style = "nochanges"
		self.update(self._generate(plot_style, custom_dict))

	def update(self, *args,**kwargs):
		super().update(*args,**kwargs)
		self.update_mplrcParams()

	def _check_style_parameter(self, key):
		
		if not key in self.mpl.rcParams.keys():
			raise KeyError
		# Ignore a list of deprecated matplotlib rcParams attributes
		if key in ['text.latex.unicode', 'axes.prop_cycle', 'examples.directory']:
			return False
		else:
			return True

	def __setitem__(self, key, val):
		'''
		Overriding the __setitem__ method to avoid setting parameter key words that are not required.
		'''
		try:
			if self._check_style_parameter(key):
				dict.__setitem__(self, key, val)
		except KeyError:
			raise KeyError('%s is not a valid matplotlib plotting parameter.' % (key,))

	def _generate(self, style=None, custom=None):
		"""
		The purpose of this method is to generate dictionaries using the ``style``, and
		allow ``custom`` keys to modify what is specified in the custom styles
		"""
		import copy

		if not isinstance(style, str):
			return {}
		if style in plot_styles.keys():
			newstyle = copy.copy(plot_styles[style])
			if isinstance(custom, dict):
				newstyle.update(custom)

		# allow 'pt' followed by a int or float number for the size of the font for all font related plot style prameters
		elif style.startswith('pt'):
			newstyle = copy.copy(ccgpaper)
			idx = style.rfind('t')
			newstyle.update(ccgpaper.fromkeys(fontparms, float(style[idx + 1:])))
			if isinstance(custom, dict):
				newstyle.update(custom)
		else:
			tempstyles = list(plot_styles.keys())
			tempstyles.append('pt##')
			raise ValueError("The style '%s' is not available in pygeostat. Please use one of the"
							 " following:\n%s" % (style, tempstyles))
		# make sure the keys are permissible mpl keys
		for key in newstyle:
			if key not in mpl.rcParams:
				raise KeyError("ERROR: {} not a valid rcParams key!".format(key))
		return newstyle
		

	def update_mplrcParams(self, style=None, custom=None):
		"""
		Update the current mpl rcParams by drawing from the set of pygeostat plotting styles,
		with optional custom parameters.

		Parameters:
			style (str): One of the pygeostat default plotting styles
			custom (dict): dictionary of mpl rcParams and values to set
		"""
		from matplotlib import cbook
		import warnings

		if Parameters['config.ignore_mpl_warnings']:
			warnings.filterwarnings("ignore", category=cbook.MatplotlibDeprecationWarning)
		else:
			warnings.filterwarnings('default', category=cbook.MatplotlibDeprecationWarning)

		# Required in order to restore mlp rcParams
		self._old_rcParams = self.mpl.rcParams.copy()
		if style in (None, False):
			if custom in (None, False):
				self.mpl.rcParams.update(self)
			else:
				tempkwargs = self.copy()
				tempkwargs.update(custom)
				self.mpl.rcParams.update(tempkwargs)
		else:
			tempkwargs = self._generate(style, custom)
			self.mpl.rcParams.update(tempkwargs)

	
	def restore_defaults(self):
		"""
		Restore the matplotlib defaults for this session
		"""
		self.update(self._default_dict)

	def restore_mplrcParams(self):
		""" Restores the `old` rcParams saved prior to the style update """
		mpl.rcParams.update(self._old_rcParams)


	def get_systemdefault(self, confirm_autoload=False, verbose = True):
		"""
		Get the PygeostatParams stored in the system defaults in ``%USER%/.PygeostatParams``. ``confirm_autoload``
		is passed only on the very first call on module init.
		"""
		if os.path.isfile(_params_cfg_file):
			newdict = _parse_pardict(_params_cfg_file, section=self._section_name)
			
			if newdict: # If dictionary is not empty
				if confirm_autoload:
					if Parameters.get("config.autoload.plot_style", False):
						self.update(newdict)
						if Parameters.get("config.verbose", False):
							print("Loading default Pygeostat Parameters from {}".format(_params_cfg_file))
				else:
					self.update(newdict)
					if Parameters.get("config.verbose", False):
						print("Loading default Pygeostat Parameters from {}".format(_params_cfg_file))

#--------------------Initilizations-------------------
#-----------------------------------------------------
# 1) Parameters comes first
def _parameters_initialize():
	ret = Parameters([(key, default) for key, (default, _) in
					six.iteritems(default_parameters)])
	ret.get_systemdefault(confirm_autoload=True)
	return ret

Parameters = _parameters_initialize()

# 2) Setting the plot styles 
def _plot_style_initilizer():
	ret = PlotStyle('ccgpaper')
	ret.get_systemdefault(confirm_autoload=True)
	return ret

PlotStyle = _plot_style_initilizer()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# public
"""imagegrid.py: Provides a means to easily generate grids of plots"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

from . set_style import set_plot_style


@set_plot_style
def subplots(nrows=1, ncols=1, figsize=None, axes_pad=(0.9, 0.4), aspect=True, label_mode='L',
             cbar_mode=None, cbar_location='right', cbar_pad=.03, cbar_size='3%', cbar_set_cax=True,
             share_all=True, fig=None, axes_class=None, rect=111):
    '''
    A wrapper of the Matplotlib ImageGrid class, providing additional ease of use for the following
    reasons:

        - The format of the arguments is modified to mimic the more commonly used
          matplotlib.pyplot.subplots
        - A matplotlib.figure.Figure no longer needs to be initiated seperately
        - Kwarg locations and defaults are modified to emphasize more commonly used
          and modified settings. Rarely used kwargs such as direction and ngrids are removed.
        - The involved import statement of ImageGrid is avoided
        - Project defaults are integrated via pygeostat.Parameters (future work)

    ImageGrid:

        "A class that creates a grid of Axes. In matplotlib, the axes location (and size) is specified
        in the normalized figure coordinates. This may not be ideal for images that needs to be
        displayed with a given aspect ratio.  For example, displaying images of a same size with some
        fixed padding between them cannot be easily done in matplotlib. ImageGrid is used in such case."

    Parameters:
        nrows(int): number of of rows in the grid
        ncols(int): number of of columns in the grid

    Optional keyword arguments (from ImageGrid):

        ================  ========  =========================================
        Keyword           Default   Description
        ================  ========  =========================================
        axes_pad          0.02      float| pad between axes given in inches
                                    or tuple-like of floats,
                                    (horizontal padding, vertical padding)
        aspect            True      [ True | False ] If True, each the length
                                    of the x and y axes are based on their
                                    absolute range. Generally used for plots
                                    such as maps, but should be False for
                                    plots such as variograms.
        label_mode        "L"       [ "L" | "1" | "all" ]
        cbar_mode         None      [ "each" | "single" | "edge" ]
        cbar_location     "right"   [ "left" | "right" | "bottom" | "top" ]
        cbar_pad          None
        cbar_size         "5%"
        cbar_set_cax      True      [ True | False ] if True, each axes in
                                    the grid has a cax attribute that is bind to
                                    associated cbar_axes.
        share_all         False     [ True | False ]
        axes_class        None      a type object which must be a subclass
                                    of axes_grid's subclass of
                                    :class:`~matplotlib.axes.Axes`
        ================  ========  =========================================

    Keyword Arguments:
        figure(matplotlib.figure.Figure): a figure is created if one is not
            provided (based on rcParams['figure.figsize'] if None)
        figsize(matplotlib.figure.Figure): size of the figure, if one must
            be created. Based on rcParams['plotting.figsize'] if None
        rect(int): "[left, bottom, width, height]* (in:class:`
            ~matplotlib.figure.Figure` coordinates) or the subplot position code
            (e.g., "121"). The default setting occupies the entire figure.

    Returns:
        (fig, axes): mirroring the output of matplotlib.subplots()

    .. seealso::
        :func:`Subplots Clean <pygeostat.plotting.subplots.subplots_clean>`


    **Example:**

    Plot 4 realizations in each panel with a common colorbar.

    .. plot::

        import pygeostat as gs

        # Initialize gs.subplots
        fig, axes = gs.subplots(2, 2, figsize=(8, 8), cbar_mode='single')

        # Default grid definition
        gs.Parameters['data.griddef'] = gs.GridDef('120 5 10\\n110 1205 10\\n1 0.5 1.0')

        # Iterate over the axes/realizations
        for i, ax in enumerate(axes):
            sim = gs.ExampleData('grid2d_surf_real'+str(i+1))
            gs.slice_plot(sim, var='Top Elevation', ax=ax, vlim=(375, 385),
                        title='Realization '+str(i+1))


    .. codeauthor:: pygeostat development team 2018-04-05
    '''
    from mpl_toolkits.axes_grid1 import ImageGrid
    from matplotlib.pyplot import figure
    if fig is None:
        fig = figure(figsize=figsize)
    if rect is None:
        rect = 111
    axes = ImageGrid(fig, rect, (nrows, ncols),
                     axes_pad=axes_pad, share_all=share_all,
                     aspect=aspect, label_mode=label_mode,
                     cbar_mode=cbar_mode, cbar_location=cbar_location,
                     cbar_pad=cbar_pad, cbar_size=cbar_size,
                     cbar_set_cax=cbar_set_cax, axes_class=axes_class)
    return (fig, axes)


def subplots_clean(axes, nused, ncols=None):
    '''Remove unused axes. Add labels where space is made available due to
       the removed axes.

       Parameters:
        axes (list of axes from gs.subplots or plt.subplots): where the
            number of axes exceeds the number used (nused)
        nused (int):number of used axes, where it assumed that axes[:nused]
            are used.
        ncols(int): number of of columns in the grid, which is required to
            add labels.
    '''
    nax = len(axes)
    if nused >= nax:
        raise ValueError('nused >= len(axes)! nothing can be done!')
    # Remove the unrequired axes
    for ax in axes[nused:]:
        ax.remove()
    if ncols is not None:
        if nax - nused > ncols:
            raise ValueError('len(axes) - nused > ncols! use a smaller grid!')
    # Add xlabels where required
    for i, ax in enumerate(axes[:nused]):
        if i + ncols >= nused:
            ax.axis['bottom'].toggle(ticklabels=True, label=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Varplot wrapper to allow variogram reproduction plotting"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

import matplotlib as mpl
import matplotlib.pyplot as plt

from . set_style import set_plot_style
from .. pygeostat_parameters import Parameters

@set_plot_style
def variogram_plot_simulation(simulated_data, simulated_var_id, reference_data=False, reference_var_id=1, sill=1, ax=None, figsize=None,
                            xlim=None, ylim=None, trimylim=True, title=None, variable=None, xlabel=None,
                            unit=None, ylabel=None, refclr=None, simclr=None, alpha=None, ls='-', lw=1,
                            lw_real=None, legend_label=None, plot_style=None, custom_style=None, output_file=None,
                            out_kws=None, **kwargs):
    """
    variogram_plot_simulation provides a means of checking variogram reproduction of simulation realizations to the
    input variogram model. A variogram reference system is used to know which variograms to plot
    from a output GSLIB varsim file. Please refer to :func:`gs.get_uniquevarids()
    <pygeostat.plotting.variogram_utils.get_uniquevarids>` for information to understand how it operates.

    The function uses pygeostats :func:`gs.variogram_plot() <pygeostat.plotting.variogram_plot.variogram_plot>` to plot
    both simulation, and if present, reference variograms.

    The only required parameters are ``simulated_data`` and ``simulated_var_id``. The function only accepts
    ``simulated_data`` as a pandas dataframe. ``simulated_var_id`` is the corresponding  ID value derived from
    :func:`gs.get_uniquevarids() <pygeostat.plotting.variogram_utils.get_uniquevarids>`, to the function
    knows which  variograms to plot. All other parameters are optional or have default values. The
    line width of the simulation plots are hard coded to be half of the size input with the ``lw``
    parameter.

    Please review the documentation of the :func:`gs.set_style()
    <pygeostat.plotting.set_style.set_style>` and :func:`gs.export_image()
    <pygeostat.plotting.export_image.export_image>` functions for details on their parameters so that
    their use in this function can be understood.

    Parameters:
        simulated_data (pd.DataFrame): Dataframe containing the following columns: 'Variogram Index',
            'Variogram Number', 'Calculation Azimuth', and 'Calculation Dip'. Variogram distance
            data is also needed as must have one of the following column headers: 'h', 'Lag
            Distance', or 'Distance'
        simulated_var_id (int): A 'Variogram ID' derived from :func:`gs.get_uniquevarids()
            <pygeostat.plotting.variogram_utils.get_uniquevarids>`
        reference_data (pd.DataFrame or bool): Dataframe containing the reference variogram value,
            variogram distance, and variogram index (if required) data as columns. The dataframe
            must contain the correct column IDs. The column header containing the variogram
            distance can be: 'h', 'Lag Distance', or 'Distance.' The column header containing the
            variogram values can be: 'vario', 'Variogram Value', or 'Variogram'. If reference data
            does not need to be plotted, set ``reference_data`` to ``False``
        reference_var_id (int): Point to which reference variogram you would like to plot if there are
            multiple variogram within your reference dataframe. The dataframe must contain the
            correct column ID. The column header containing the variogram index values can be:
            'Variogram Index' or 'Index'
        sill (float): Value to plot a horizontal line representing the variograms sill
        ax (mpl.axis): Matplotlib axis to plot the figure
        figsize (tuple): Figure size (width, height)
        xlim (float tuple): Minimum and maximum limits of data along the x axis
        ylim (float tuple): Minimum and maximum limits of data along the y axis
        trimylim (bool): Indicate if realization plots should cease plotting once they pass the
            limits specified by ``ylim``.
        title (str): Title for the plot. Can use ``dirs`` from :func:`gs.get_uniquevarids()
            <pygeostat.plotting.variogram_utils.get_uniquevarids>` if direction data is desired in the
            title.
        variable (str): By default, titles are generated based on the data provided, which include
            the variable number and direction information. If you would like to keep the direction
            data but update the variogram name manually, use this parameter to give that variable
            name
        xlabel (str): X-axis label
        yalabl (str): Y-axis label
        unit (str): Distance units used for lag distance. Only used if the keyword parameter
            ``xlabel`` is left to its default value of ``None``.
        refclr (str): Any Matplotlib color for the reference variogram
        simclr (str): Any Matplotlib color for the realization variograms
        alpha (float): Transparency for realization variograms (0 = Transparent, 1 = Opaque)
        ls (float): A valid Matplotlib line style for both reference and realization variograms
        lw (float): Line width in points. The width provided in this parameter is used for the
            reference variogram.
        lw_real (float): Line width in points. If no value is passed, half the value of ``lw`` is
            used for the realization variograms
        legend_label (str or bool): A string containing the label that will be attached to the
            reference variogram if it exists. If the value `True` is passed, only the realizations
            will have a label.
        plot_style (str): Use a predefined set of matplotlib plotting parameters as specified by
            :class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
            to turn it off
        custom_style (dict): Alter some of the predefined parameters in the ``plot_style`` selected.
        output_file (str): Output figure file name and location
        out_kws (dict): Optional dictionary of permissible keyword arguments to pass to
            :func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`
        **kwargs: Optional permissible keyword arguments to pass to :func:`gs.variogram_plot()
            <pygeostat.plotting.variogram_plot.variogram_plot>` and by extension, matplotlib's plot function if the
            keyword passed is not used by :func:`gs.variogram_plot() <pygeostat.plotting.variogram_plot.variogram_plot>`

    Returns:
        ax (ax): matplotlib Axes object with the variogram

    Examples:
        Import a variogram model manually into python using pygeostat functions:

        >>> #Import the reference variogram models into python
        >>> var1 = gs.VargModel(vargstr='''3 0
        >>> 1 0.46 0 0 0
        >>> 16 16 16
        >>> 1 0.06 0 0 0
        >>> 32 32 32
        >>> 1 0.48 0 0 0
        >>> 64 64 64''')
        >>> #Generate a dataframe that has lag distances values for the above model
        >>> model1 = var1.model(azm=0, dip=0, nlags=200, lagdist=0.5, returnstr=False)

        Load output from varsim:

        >>> varsimulated_data = gs.DataFile('varsim_reals.out', readfl=True)

        Look at what variograms are within the variogram data:

        >>> gs.get_uniquevarids(varsimulated_data.data)
        Variogram ID: 1 ... Variable: 1, Azimuth 90, Dip 0
        Variogram ID: 2 ... Variable: 2, Azimuth 90, Dip 0
        Variogram ID: 3 ... Variable: 3, Azimuth 90, Dip 0

        A simple call using the variogram model loaded and the set of realizations associated with
        Variogram ID 1:

        >>> gs.variogram_plot_simulation(simulated_data=varsimulated_data.data, simulated_var_id=1, reference_data=model1)

        .. image:: ./figures/variogram_plot_simulation1_150.png

        |

        Use a custom variogram title, and fix the x-axis limits:

        >>> gs.variogram_plot_simulation(simulated_data=varsimulated_data.data, simulated_var_id=1, reference_data=model1, xlim=(0,150),
        ...              title='PVar1 Variogram Reproduction')

        .. image:: ./figures/variogram_plot_simulation2_150.png

        |

        To plot all the variables at once, we first need a dictionary that has model data loaded
        within it. Create a dictionary using the already loaded model and files generated from
        varcalc for the other variables:

        >>> models={}
        >>> models[1] = model1
        >>> models[2] = gs.DataFile('varsim_model_var2.out', readfl=True).data
        >>> models[3] = gs.DataFile('varsim_model_var3.out', readfl=True).data

        Now loop through the variables, use a variable name to update the title produced:

        >>> variables=['PVar1', 'PVar2', 'PVar3']
        >>> for i in range(1,4):
        >>>     var=variables[i-1]
        >>>     gs.variogram_plot_simulation(simulated_data=varsimulated_data.data, simulated_var_id=i, reference_data=models[i], variable=var,
        ...                  xlim=(0,150))

        .. image:: ./figures/variogram_plot_simulation3_150.png
        .. image:: ./figures/variogram_plot_simulation4_150.png
        .. image:: ./figures/variogram_plot_simulation5_150.png

    """
    import pygeostat as gs
    from .variogram_utils import get_uniquevarids
    from . utils import format_plot, setup_plot
    from .export_image import export_image

    # Handle dictionary defaults
    if out_kws is None:
        out_kws = dict()
    # Get the enumerate data for the datafiles that exist
    simulated_var_ids, simdirs = get_uniquevarids(simulated_data, mode='ref', source='varsim')
    if not isinstance(reference_data, bool):
        if 'Variogram Number' in reference_data.columns:
            reference_var_ids, _ = get_uniquevarids(reference_data, mode='ref')
            refindex = reference_var_ids[reference_var_id]
        else:
            refindex = None
    # Set-up plot if no axis is supplied
    fig, ax, cax = setup_plot(ax, figsize=figsize, aspect =False)

    #  Set-up some plotting defaults
    if lw_real is None:
        lw_real = lw / 2
    if ylim is None and sill == 1:
        ylim = (0, 1.2)
    # Plot simulation variograms
    if trimylim:
        if isinstance(ylim, tuple):
            simulated_data = gs.trimylim(simulated_data, ylim[1])
    if simclr is None:
        simclr = Parameters['plotting.variogram_plot_simulation.simclr']
    if alpha is None:
        alpha = Parameters['plotting.variogram_plot_simulation.alpha']
    for i, index in enumerate(simulated_var_ids[simulated_var_id]):
        if legend_label:
            if i == (len(simulated_var_ids[simulated_var_id]) - 1):
                label = 'Realizations'
            else:
                label = False
        else:
            label = False
        gs.variogram_plot(data=simulated_data, index=index, sill=False, experimental=False, ax=ax, color=simclr,
                  ylabel=False, xlabel=False, title=False, lw=lw_real, alpha=alpha, label=label,
                  plot_style=None, **kwargs)
    # Plot the reference variogram
    if refclr is None:
        refclr = Parameters['plotting.variogram_plot_simulation.refclr']
    if not isinstance(reference_data, bool):
        if legend_label is True:
            legend_label = False
        gs.variogram_plot(data=reference_data, sill=False, experimental=False, index=refindex, ax=ax, color=refclr,
                  ylabel=False, xlabel=False, title=False, lw=lw, label=legend_label,
                  plot_style=None,
                  **kwargs)
    # Plot figure and axis labels
    if unit is None:
        unit = Parameters['plotting.unit']
    if unit is None or unit == '':
        unit = ''
    else:
        unit = ' ({})'.format(unit)
    if xlabel is None:
        ax.set_xlabel(Parameters['plotting.lagname'] + unit)
    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel is None:
        ax.set_ylabel(r'$\gamma$   ', fontsize=mpl.rcParams['font.size'] *
                      Parameters['plotting.gammasize'], rotation=0)
    if ylabel:
        ax.set_ylabel(ylabel)
    if title is None:
        ndirs = len(simdirs)
        if variable is None:
            var = "Variable %s" % int((simulated_var_id / ndirs) + 0.5)
        else:
            var = variable
        direction = (simulated_var_id % ndirs) - 1
        if direction < 0:
            direction = ndirs - 1
        azm = simdirs[direction][0]
        dip = simdirs[direction][1]
        ax.set_title("%s (Azm: %s°, Dip: %s°)" % (var, azm, dip))
    if title:
        ax.set_title(title)
    # Set plot limits
    if not isinstance(sill, bool):
        ax.plot(ax.get_xlim(), (sill, sill), 'k-', ls='-', lw=1.5)
    if xlim:
        ax.set_xlim(xlim)
    if ylim:
        ax.set_ylim(ylim)
    if legend_label:
        ax.legend()
    # Export figure
    if output_file or ('pdfpages' in out_kws):
        export_image(output_file, **out_kws)
    #  Return axis
    return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Function that alters the matplotlib rc parameters as desired with some predefined styles"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
from functools import wraps
import matplotlib as mpl
from matplotlib import pyplot as plt
from ..pygeostat_parameters import PlotStyle


def set_plot_style(plot_function):
    """
    Applies pygeostat plotting style before plotting function is called based on the provided style name and custom dictionary then
    retore it back to the original plot style.

    **Example:**

    .. code-block:: Python

        import pygeostat as gs

        @gs.set_plot_style
        def coolnewplot(data, griddef, **otherkwargs, plot_style=None, custom_style=None):
            # plot some things
            # apply cool formatting
            return things
    """
    @wraps(plot_function) # Using the wraps for invoking update_wrapper(). This keeps the name and doc string of the original function
    def wrapper_plot_function(*args, **kwargs):
        # set the passed style
        PlotStyle.update_mplrcParams(kwargs.get("plot_style", None), 
                                       kwargs.get("custom_style", None))
        # call the plotting function
        returnargs = plot_function(*args, **kwargs)
        # plt.show()
        # set the style back to whatever the default was
        PlotStyle.restore_mplrcParams()
        return returnargs
    
    return wrapper_plot_function



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Plot multi dimensional scaling (MDS) coordinates.
"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

from . set_style import set_plot_style

@set_plot_style
def mds_plot(data, variables=None, enumerate_vars=True, ax=None, figsize=(3, 3), title=None, aspect = None,
			cmap='viridis', s=15, varlabels=None, plot_style=None, custom_style=None, output_file=None, out_kws=None,
			**kwargs):
	"""
	Python implementation of the MDS coordinates ploted by the CCG program corrmat when set to
	'ordination' mode.

	MDS coordinates are calculate from the correlation matrix.

	.. seealso::

		1. Deutsch, M. V, & Deutsch, C. V. (2013). A Program to Calculate and Display Correlation
		   Matrices with Relevant Analysis. Edmonton AB. Retrieved from http://www.ccgalberta.com

	Parameters:
		data: Tidy (long-form) 2-D data where each column is a variable and each row is an
			observation. A pandas dataframe or numpy array may be passed.
		variables (list): Variables from the pd.DataFrame passed with ``data`` to calculate
			coordinates for
		enumerate_vars (bool): Indicate if the scatter points should be annotated with an
			enumerator rather then the variable name. A reference legend is placed to the right
			of the plot if set to ``True``
		ax (mpl.axis): Matplotlib axis to plot the figure on
		figsize (tuple): Figure size (width, height) in inches
		title (str): Title for the plot
		cmap (str): A matplotlib colormap object or a registered matplotlib or pygeostat colormap
			name.
		s (float): Size of location map markers
		plot_style (str): Use a predefined set of matplotlib plotting parameters as specified by
			:class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
			to turn it off
		custom_style (dict): Alter some of the predefined parameters in the ``plot_style`` selected.
		output_file (str): Output figure file name and location
		out_kws (dict): Optional dictionary of permissible keyword arguments to pass to
			:func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`
		**kwargs: Optional dictionary of permissible keyword arguments to pass to the scatter plot

	Examples:
		A simple call:

		>>> gs.mds_plot(data.data)

		.. image:: ./figures/mds_plot1.png

	"""

	from .cmaps import avail_cmaps
	from .utils import get_contcbarargs, setup_plot, smart_annotate, get_cmap
	from .export_image import export_image
	from .. multivariate.utils import mds
	import numpy as np

   
	# Default dictionary
	if not out_kws:
		out_kws = dict()
	# Function set-up and defaults
	if variables is None:
		try:
			variables = list(data.columns)
		except:
			variables = list(np.arange(data.data.shape[1]))
	if enumerate_vars is True:
		annots = list(range(1, len(variables) + 1))
	else:
		annots = variables
	if cmap in avail_cmaps:
		cmap = get_cmap(cmap)
	# Calculate the MDS coordinates
	coords = mds(data, variables)
	xcoord = coords['X1']
	ycoord = coords['X2']
	zcoord = coords['X3']
	# Make the figure object
	fig, ax, cax = setup_plot(ax, cax=None, cbar=True, figsize=figsize, **kwargs)
	ax.set_aspect(aspect)
	# Plot the scatter plot using the MDS coordinates
	vlim, ticklocs, ticklabels = get_contcbarargs(zcoord, 2, None)
	plot = ax.scatter(xcoord, ycoord, c=zcoord, cmap=cmap, vmin=vlim[0], vmax=vlim[1], s=s,
					  alpha=0.8)
	# Plot scatter plot outlines to better show overlapping
	ax.scatter(xcoord, ycoord, c='none', s=s)
	# Adjust the annotations of each point
	smart_annotate(ax, xcoord, ycoord, annots)
	# Set-up the rest of the plots
	ax.spines['top'].set_visible(False)
	ax.spines['right'].set_visible(False)
	if title is not None:
		ax.set_title(title)
	ax.tick_params(axis='both', pad=2)
	ax.set_xlabel('X1')
	ax.set_ylabel('X2')
	# Configure the color bar
	cbar = fig.colorbar(plot, cax=cax, ticks=ticklocs)
	cbar.ax.set_yticklabels(ticklabels, ha='left')
	cbar.ax.tick_params(axis='y', pad=2)
	cbar.set_label('X3', ha='center', va='top', labelpad=2)
	# Write the legend if the variables are enumerated
	if enumerate_vars:
		renderer = fig.canvas.get_renderer()
		fig.draw(renderer)
		pad = len('%s' % len(variables))
		colheight = cbar.ax.yaxis.get_tightbbox(renderer).height
		txt = cbar.ax.yaxis.label
		txtx = ax.transAxes.inverted().transform((txt.get_window_extent(renderer).xmax,
												  txt.get_window_extent(renderer).xmax))[0] + 0.02
		legend = ''
		nlines = 0
		if varlabels is None:
			varlabels = variables
		for i, var in enumerate(varlabels):
			legend = legend + '{number:{pad}d}: {var}   \n'.format(pad=pad, number=(i + 1),
																   var=var)
			if nlines == 0:
				txt = ax.text(txtx, 1, legend + '\nTest', ha='left', va='top',
							  transform=ax.transAxes)
				height = txt.get_window_extent(renderer).height
				txt.remove()
				if height > colheight:
					nlines = (i + 1)
			if (nlines > 0) and (((i % nlines) == (nlines - 1)) or
								 (i == len(variables) - 1)):
				txt = ax.text(txtx, 1, legend, ha='left', va='top', transform=ax.transAxes)
				txtx = ax.transAxes.inverted().transform((txt.get_window_extent(renderer).xmax,
													  txt.get_window_extent(renderer).xmax))[0]
				legend = ''
			elif var == varlabels[-1]:
				txt = ax.text(txtx, 1, legend, ha='left', va='top', transform=ax.transAxes)
	# Export figure
	if output_file or ('pdfpages' in out_kws):
		export_image(output_file, **out_kws)

	return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''programs.py: Contains methods for running GSLIB and CCG programs'''
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
import numpy as np

from ..utility.logging import log_progress, printerr
from .. pygeostat_parameters import Parameters


def runparallel(gslibprogram, kwargslist, nprocess=None, mute=False, progressbar=False):
    '''
    Run a set of gslib program calls in parallel

    Parameters:
        gslibprogram (Program): name of GSLIB/CCG program to run
        kwargslist (list of dictionaries): list of keyword arguments which will be used to call
            gslibprogram.run(kwarg)
        nprocess (int): number of threads to spawn.  Drawn from Parameters['config.nprocess'] if None.

    Examples:

        Setting up the calling parameters. This example is based off the example used in
        :func:`gs.Program() <pygeostat.programs.programs.Program>`

        >>> callpars = []
        >>> # For each variable we want to run in parallel, assemble a dictionary of parameters and
        ... # append to callpars
        >>> for variable in ['Bitumen','Fines','Chlorides']:
        >>>     # Establish the parameter file for this variable
        >>>     mypars = {'datafl':datafl.flname,
        ...               'varcol':datafl.gscol(variable),
        ...               'varname':variable,
        ...               'outfl':'histplt_'+variable+'.ps'}
        >>>    # Assemble the arguments for the GSLIB call and add the arguments to the list of calls
        >>>    callpars.append({'parstr':histpltpar.format(**mypars),
        ...                     'parfile':'histplt_'+variable+'.par',
        ...                     'testfilename':datafl.flname})

        Now run in parallel

        >>> histplt = gs.Program(program='histplt', parfile='histplt.par')
        >>> gs.runparallel(histplt, callpars)
        '''
    import multiprocessing as mp
    from os import remove

    progstr = gslibprogram.program
    if '/' in progstr:
        progstr = progstr[progstr.rfind('/') + 1:]
    if '.' in progstr:
        progstr = progstr[:progstr.rfind('.')]
    if not mute:
        print('Creating parallel processes')
    # Create a pool for processing, calling with None to use all CPUs
    if nprocess is None:
        nprocess = Parameters['config.nprocess']
    pool = mp.Pool(processes=nprocess)
    # Before executing, check to see if a parfile is present in each
    # kwarg. If not present then append a parfile that is unique for
    # each execution
    temppars = []
    for i, kwargs in enumerate(kwargslist):
        if 'parfile' not in kwargs.keys():
            temppar = progstr + "_" + temp_gslib_file("par")
            kwargslist[i]['parfile'] = temppar
            temppars.append(temppar)
    if progressbar:
        pbar = log_progress(range(len(kwargslist)))

        def pbar_update(*args, **kwargs):
            pbar.update()
    else:
        pbar_update = None
    # Now call in parallel for each kwargs
    for kwargs in kwargslist:
        # Add on the GSLIB call to execute asynchronously
        pool.apply_async(gslibprogram.run, (), kwargs, callback=pbar_update)
    # Close the list of processes for the pool
    if not mute:
        print('Pool assembled, asynchronously processing')
    pool.close()
    # "Join" the list of processes to execute and wait for completion
    pool.join()
    if not mute:
        print('Asynchronous execution finished.')
    if progressbar:
        pbar.close()
    # Remove the temporary parameter files if any were created
    if len(temppars) > 0:
        for temppar in temppars:
            remove(temppar)


class GennyWithAMemry:
    """
    A random seed generator object, callable, to replace gs.rseed, to generate more unique
    random seeds
    """

    def __init__(self, seed=None):
        self.seed = seed
        self._prng = np.random.RandomState(self.seed)
        self._rnd_list = []

    def _grnd(self, nseeds=100000):
        rndstart = self._prng.randint(0, 232135)
        dat = np.arange(rndstart, rndstart + nseeds)
        self._prng.shuffle(dat)
        return dat.tolist()

    def __call__(self, prng=None):
        '''Returns a ACORNI (GSLIB-suitable) random number seed'''
        if prng is not None:
            if isinstance(prng, int):
                self._prng = np.random.RandomState(prng)
            else:
                self._prng = prng
            self._rnd_list = []
        if len(self._rnd_list) == 0:
            self._rnd_list = self._grnd()
        return self._rnd_list.pop(0)


# this function replaces the rseed function to generate random seeds that are guaranteed unique
# for up to 100000 calls to the function. For more unique random seeds, use `rseed_list`
rseed = GennyWithAMemry()


def rseed_list(nseeds, seed=None):
    '''
    Returns a list of ACORNI (GSLIB-suitable) random number seeds. A initial seed can be passed
    ensureing the same list of seeds is returned to a script that is rerun.

    Parameters:
        nseeds (int): Number of seeds to return
        seed (int): Initialization seed

    Returns:
        seeds (list): List of random number seeds

    '''
    return GennyWithAMemry(seed)._grnd(nseeds)


def parallel_function(function, arglist=None, kwarglist=None, nprocess=None, returnvals=False,
                      progressbar=False):
    """
    Quickly parallelize a function with a set of arguments or keyword arguments. If the function
    returns something (as oppose to writes out values to files), set returnvals=True to get the
    dictionary that can be used to collect the results.

    Parameters:
        function (func): a callable function **DEFINED IN A .py FILE**. The function must be imported
            from a module since it has to be pickled to be parallelized. Defining the function in
            the jupyter notebook doesnt seem to work.
        arglist (list or tuples): a list of **tuple** arguments to pass to the function, see
            examples
        kwarglist (list): a list of keyword dictionaries to pass to the functions, i.e.
            [{'arg1': value, 'arg2': value}, {'arg1': value, 'arg2': value}, etc]
        nprocess (int): the number of parallel processes to run. Drawn from Parameters['config.nprocess']
            if None.
        returnvals (bool): if the function returns something, collect it in a dictionary, you can
            use the .get() method of the parallel result to collect the required data.

    Returns:
        res (dict): optionally return a dictionary of parallel processing results


    Usage:

    For a function that takes a single argument, setup arglist in this way:

        >>> arglist = []
        >>> arglist.append((arg,))

    OR:

        >>> arglist = [(arg,) for arg in range(nparallel)]

    Alternatively the argument tuple may take several arguments:

        >>> arglist = [(args, for, function), (args, for, function), (args, for, function)]

    Detailed usage:

        >>> arglist = []
        >>> for sr in series:
        ...     arglist.append(('keyout.out', rbfpath + 'keyout%s.out' % sr, griddef,
        ...                     griddefs[sr], [3]))
        >>> gs.parallel_function(rm.changegrid, arglist=arglist)

    OR:

        >>> kwarglist = []
        >>> for sr in series:
        ...     kwarglist.append({'infl': 'keyout.out',
        ...                       'outfl': rbfpath + 'keyout%s.out' % sr,
        ...                       'ingrid': griddef,
        ...                       'outgrid': griddefs[sr],
        ...                       'avmethods': [3]
        ...                       })
        >>> gs.parallel_function(rm.changegrid, kwarglist=kwarglist)
    """
    # some minor checks
    if arglist is None and kwarglist is None:
        raise ValueError('Must pass either lists of arglist or a list of kwarglist')
    if arglist is not None and kwarglist is not None:
        raise ValueError('Pass arglist OR kwarglist.. not both!')
    # start the parallel pool
    import multiprocessing as mp
    if nprocess is None:
        nprocess = Parameters['config.nprocess']
    if progressbar:
        pbar = log_progress(range(len(kwarglist if arglist is None else arglist)))

        def pbar_update(*args, **kwargs):
            pbar.update()
    else:
        pbar_update = None
    pool = mp.Pool(processes=nprocess)
    res = {}
    if arglist is not None:
        if not all([isinstance(item, tuple) for item in arglist]):
            printerr('all arguments in `arglist` must be of type `tuple`. If only 1 argument is '
                     'being passed to the parallel function, use: [(arg,)] (see docs). Attempting '
                     'to automagically convert the arguments to tuples ...')
        try:
            arglist = [item if isinstance(item, tuple) else tuple([item]) for item in arglist]
        except TypeError:
            printerr('items in `arglist` cannot be converted to type `tuple` automatically')
        # call the parallel function with the arguments
        for i, args in enumerate(arglist):
            res[i] = pool.apply_async(function, args, callback=pbar_update)
    else:
        # call the parallel function with the kwargs
        for i, kwargs in enumerate(kwarglist):
            res[i] = pool.apply_async(function, (), kwargs, callback=pbar_update)
    pool.close()
    pool.join()
    if progressbar:
        pbar.close()
    if returnvals:
        return res


def parstr_kwargs(parstr, fmt='pars'):
    """
    Print a formatted list of kwargs found in the parfile. Tested and working for the {} style
    string formatting, found in the gamsim_ave parfile found below this function (for exampel).
    This is mostly used for being lazy and not writing out the kwargs you just entered into the
    parfile.... can also be helpful if you define the parfile elsewhere and you cant remember what
    kwargs you setup !

    Parameters
    ----------
    parstr: str
        the par string with the {} formatted parameters
    fmt: str
        the output format, permissible arguments are `pars` or `dict`

    Examples
    --------
    Get the formatted parfile

        >>> parstr = '''          Parameters for GAMSIM_AVE
        >>>             *************************
        >>> START OF PARAMETERS:
        >>> {lithfl}        -file with lithology information
        >>> {lithcol}   {lithcode}                        -   lithology column (0=not used), code
        >>> {datafl}       -file with data
        >>> ....
        >>> '''

        >>> print_parfile_kwargs(parstr)
        ... lithfl=, lithcol=, lithcode=, datafl=, ...

    """
    from string import Formatter
    if fmt.lower() == 'dict':
        print(', '.join(["'%s': " % s[1] for s in Formatter().parse(parstr)][:-1]))
    else:
        print(', '.join(['%s=' % s[1] for s in Formatter().parse(parstr)][:-1]))


def dedent_parstr(indented_parstr):
    """
    Remove leading indents on each line from a parameter file. This is automatically called by
    :func:`Program.run() <pygeostat.programs.programs.Program.run>` so that parfiles may be tabbed to permit better
    structuring of python code.

    Examples
    --------
    An un-tabbed parstr:

        >>> parstr = '''          Parameters for GAMSIM_AVE
        >>>             *************************
        >>> START OF PARAMETERS:
        >>> {lithfl}        -file with lithology information
        >>> {lithcol}   {lithcode}                        -   lithology column (0=not used), code
        >>> {datafl}       -file with data
        >>> ....
        >>> '''

    A tabbed parstr:

        >>> parstr = '''          Parameters for GAMSIM_AVE
        >>>                 *************************
        >>>     START OF PARAMETERS:
        >>>     {lithfl}        -file with lithology information
        >>>     {lithcol}   {lithcode}                        -   lithology column (0=not used), code
        >>>     {datafl}       -file with data
        >>>     ....
        >>> '''
    """
    from textwrap import dedent
    dedented_parstr = ''
    for line in indented_parstr.splitlines():
        dedented_parstr += dedent(line) + '\n'
    return dedented_parstr


def temp_gslib_file(fltype='gslib'):
    """
    Generate unique temporary file names with different extensions as indicated by the ``fltype``
    parameter.

    Parameters
    ----------
        fltype: str
            one of "gslib", "par", "h5"
    """
    import tempfile
    import os
    cpath = os.path.relpath(os.getcwd())
    if fltype.lower() == 'gslib':
        return os.path.relpath(tempfile.NamedTemporaryFile(dir=cpath).name + '.out')
    elif fltype.lower() == 'par' or fltype.lower() == 'parfile':
        return os.path.relpath(tempfile.NamedTemporaryFile(dir=cpath).name + '.par')
    elif fltype.lower() == 'h5':
        return os.path.relpath(tempfile.NamedTemporaryFile(dir=cpath).name + '.h5')


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Function to export a matplotlib figure object as multiple files simply while minimizing
whitespace surrounding the final image"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
import os
import fileinput
import copy
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from PIL import Image
from PIL import PngImagePlugin


def export_image(output_file=None, fltype=None, pad=0.03, dpi=300, custom=None, pdfpages=None, delim=None,
              Metadata=True, **kwargs):
    """
    This function exports a figure with the specified file name and type(s) to the specified
    location. Multiple file types can be exported at once. Avoids the use of plt.tight_layout()
    which can behave odd and minimizes whitespace on the edge of figures.

    Note:
        This function is typically called within plotting functions but can be used on its own.

    Extensions are not required in the ``output_file`` argument. They will be added according to what
    ``fltype`` is set to. The default output file types are png and eps. However, if extensions
    are provided, they will be used, provided that the argument ``fltype`` is not passed. The
    ``custom`` argument  provides extra flexibility if the default settings of this function are
    not desired. If the ``custom`` functionality is the only desired output, ``fltype`` can be set
    to ``False`` to prevent additional exports.

    PS and EPS files need to have their font definitions fixed so that they will be called properly
    which is done automatically if they are used.

    Figures can also be appended to an existing ``mpl.backends.backend_pdf.PdfPages`` object passed
    with the ``pdfpages`` argument. These objects are used to created multi-page PDF documents.

    Parameters:
        output_file (str or list): Details the file location and file name in one parameter or a list of
            files to export with or without file extensions. If not file extensions are provided,
            the parameter ``fltype`` will need to be specified if its defaults are not desired
        fltype (str, list, bool): The file extension or list of extensions. See plt.savefig() docs
            for which file types are supported. Can set to ``False`` to prevent normal
            functionality when ``custom`` settings are the only desired output.
        pad (float): The amount of padding around the figure
        dpi (int): The output file resolution
        custom (dict):  Indicates a custom dpi and file extension if an odd assortment of files are
            needed
        pdfpages (mpl.object): A multi-page PDF file object created by
            ``mpl.backends.backend_pdf.PdfPages``. If a PdfPages object is passed, the figure is
            exported to it in addition to the other files if specified. Use this method to generate
            PDF files with multiple pages.
        delim (str): delimiter in the output_file str passes that indicates different types of files. set
            to `None` to ensure filenames with spaces can be used.
        kwargs: Any other permissible keyword arguments to send to plt.savefig() (e.g., )

    Examples:
        A simple call using the default settings exporting two images at 300 dpi:

        .. code-block:: python

            gs.export_image(output_file='../Figures/histplt')
            # 'histplt.png' and 'histplt.eps' are exported


        A call specifying only a png file and altering the dpi setting and setting the background
        to transparent (via \**kwargs):

        .. code-block:: python

            gs.export_image(output_file='../Figures/histplt.png', dpi=250, transparent=True)
            #'histplt.png' is exported in '../Figures/'

        A call using only the custom argument:
        
        .. code-block:: python

            gs.export_image(output_file='../Figures/histplt', fltype=False, custom={600:'png', 200:'png'})
            # 'histplt_600.png' and 'histplt_200.png' are exported in '../Figures/'

        A call using a combination of arguments:

        .. code-block:: python

            gs.export_image(output_file='../Figures/histplt', custom={600:'jpg'})
            'histplt.png' and 'histplt.eps' at 300 dip in addition to histplt_600.jpg' are exported in '../Figures/'

        A call using a more complicated combination of arguments:

        .. code-block:: python

            gs.export_image(output_file=['../Figures/png/histplt', '../Figures/eps/histplt'], custom={600:'png'})
            #'histplt.png' @ 300 dpi and 'histplt_600.png' @ 600 dpi are placed in '../Figures/png/' while 'histplt.eps' is placed in '../Figures/eps/'

        Create a PDFPages matplotlib object and save the figure to it:

        .. code-block:: python
            :linenos:

            from matplotlib.backends.backend_pdf import PdfPages
            pdfpages = PdfPages('output_file.pdf')
            gs.location_plot(data_file)
            gs.export_image(pdfpages=pdfpages)
            pdfpages.close()

    """
    def exporttompl(flname, dpi, Metadata, **kwargs):
        """
        Small function to pass information to matplotlib for exporting images for easier
        iteration
        """
        _, flext = os.path.splitext(flname)

        # Set up a pdfpage automatically so we can insert metadata
        if flext.lower() == '.pdf':
            pdffig = PdfPages(flname)
            plt.savefig(pdffig, format='pdf', bbox_inches='tight', pad_inches=pad,
                        dpi=dpi, **kwargs)
            if Metadata:
                meta_dictionary = pdffig.infodict()
                for meta in Metadata:
                    meta_dictionary[meta] = Metadata[meta]
            pdffig.close()
        else:
            plt.savefig(flname, bbox_inches='tight', pad_inches=pad, dpi=dpi, **kwargs)
            # Correct EPS or PS font definition if needed

            if flext in ['.eps', '.ps']:
                # Grab a font for now
                font = mpl.rcParams['font.family']
                for line in fileinput.FileInput(flname, inplace=1):
                    # Get the font code used in the EPS or PS file
                    if "/FontName" in line:
                        dump = copy.deepcopy(line)
                        dump = dump.split('/')
                        dump = dump[2].split(' ')
                        font = dump[0]
                    line = line.replace("/b'%s'" % font, "/%s" % font)
                    print(line, end='')
            if flext == '.png' and Metadata is not None:
                im = Image.open(flname)
                meta = PngImagePlugin.PngInfo()

                for x in Metadata:
                    meta.add_text(x, Metadata[x])
                im.save(flname, "png", pnginfo=meta)
            # if flext.lower() is in ['.jpeg', '.jpeg']:

    # --------------------------------------------------------------------------------------
    # START OF MAIN FUNCTION
    # --------------------------------------------------------------------------------------
    # Sanity check
    if not output_file and not pdfpages:
        raise ValueError("One or both of `output_file` or `pdfpages` needs to be specified.")

    # set up basic metadata info
    if Metadata is True:
        Metadata = {'Author': os.getcwd()}

    if output_file:
        # If input oufl is a string, convert it to a list
        if delim is not None:
            if isinstance(output_file, str):
                if delim in output_file:
                    return ValueError("Output files need to be delimited by a single space not a comma")
                output_file = output_file.split(delim)
        # If input fltype is a string, convert it to a list
        if isinstance(fltype, str):
            if ',' in fltype:
                return ValueError("File types need to be delimited by a single space not a comma")
            fltype = fltype.split(' ')
        # Get the extensions from output_file if they exist
        exts = []
        output_files = []
        if not isinstance(output_file, list):
            output_file = [output_file]
        for fl in output_file:
            outf, ext = os.path.splitext(fl)
            if ext is not '':
                exts.append(ext.replace(".", ""))
            output_files.append(outf)
        if len(exts) > 0 and fltype is None:
            fltype = exts
        output_file = output_files
        # Set fltype to defaults if True or None is passed and they aren't contained in output_file
        if fltype is None or (isinstance(fltype, bool) and fltype):
            fltype = ['png', 'eps']
        # Extend the output_file if needed
        if len(output_file) == 1 and fltype:
            output_file = output_file * len(fltype)
        # Export figures required through regular usage
        if not isinstance(fltype, bool) or fltype:
            for i, ext in enumerate(fltype):
                flname = '%s.%s' % (output_file[i], ext)
                exporttompl(flname, dpi, Metadata, **kwargs)
        # Export the custom files
        if custom is not None:
            for cus_dpi in custom:
                flname = '%s_%s.%s' % (output_file[0], cus_dpi, custom[cus_dpi])
                exporttompl(flname, cus_dpi, Metadata, **kwargs)

    # Export the the PDFPages object if required
    if pdfpages:
        # Try and use the variable fig if it exists (works for gs plotting functions as fig is set)
        try:
            pdfpages.savefig(fig, bbox_inches='tight', pad_inches=pad, dpi=dpi, **kwargs)
        except:
            # Grab the mpl figure if it isn't already set
            fig = plt.gcf()
            pdfpages.savefig(fig, bbox_inches='tight', pad_inches=pad, dpi=dpi, **kwargs)
